"""Simple integration tests for the AWS ELB actors."""

from nose.plugins.attrib import attr
import logging

from tornado import testing

from kingpin import utils
from kingpin.actors import exceptions
from kingpin.actors.aws import elb

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'

log = logging.getLogger(__name__)
logging.getLogger('boto').setLevel(logging.INFO)


class IntegrationELB(testing.AsyncTestCase):

    """High level ELB Actor testing.

    These tests will check two things:
    * Connection to ELB works, and instance count is correct
    * The actor continues waiting if instance count is less than expected

    Requirements:
        You have to create an ELB named kingpin-integration-test and place it
        in the specified region (default us-east-1).
        As with other tests, environment variables AWS_ACCESS_KEY_ID and
        AWS_SECRET_ACCESS_KEY are expected, and the key should have
        permissions to read ELB status.

    Note, these tests must be run in-order. The order is defined by
    their definition order in this file. Nose follows this order according
    to its documentation:

        http://nose.readthedocs.org/en/latest/writing_tests.html
    """

    integration = True

    elb_name = 'kingpin-integration-test'
    region = 'us-east-1'

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_01a_check_elb_health(self):
        actor = elb.WaitUntilHealthy(
            'Test',
            {'name': self.elb_name,
             'count': 0,
             'region': self.region})

        done = yield actor.execute()

        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test
    def integration_01b_check_elb_not_found(self):
        actor = elb.WaitUntilHealthy(
            'Test',
            {'name': 'Not-Found-ELB',
             'count': 50,
             'region': self.region})

        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield actor.execute()

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_02_wait_for_elb_health(self):
        actor = elb.WaitUntilHealthy(
            'Test',
            {'name': self.elb_name,
             'count': 1,
             'region': self.region})

        # NOTE: We are not "yielding" the execution here, but the task
        # goes on top of the IOLoop. The sleep statement below allows
        # the wait_task's actions to get executed by tornado's loop.
        wait_task = actor.execute()
        yield utils.tornado_sleep(1)

        # Expected count is 1, so this should not be done yet...
        self.assertFalse(wait_task.done())

        # Disabled this portion of the test -- as it no longer works in the
        # Tornado 4.0+ concurrent package.
        #
        # Not going to add any instances in this test, so let's abort
        # wait_task.cancel()
        # yield utils.tornado_sleep(3)
        # self.assertTrue(wait_task.done())



"""Simple integration tests for the AWS ECS actors."""

from nose.plugins.attrib import attr
import logging
import json
import os
import tempfile
import uuid
from contextlib import contextmanager

from tornado import testing

from kingpin.actors.aws import ecs

__author__ = 'Steve Mostovoy <smostovoy@nextdoor.com>'

# Generate a common UUID for this particular set of tests
UUID = uuid.uuid4().hex

log = logging.getLogger(__name__)
logging.getLogger('boto').setLevel(logging.INFO)


class IntegrationECS(testing.AsyncTestCase):

    """High level ECS Actor testing.

    This suite of tests performs the following actions:
        * Runs a series of successful and failing tasks and wait for completion
        * Runs a task without waiting
        * Starts a service with count 0
        * Changes the scale of service to 1
        * Changes the scale of service to 5 without waiting
        * Deletes service
        * Starts a service with count 1
        * Changes the scale of service to count 0
        * Deletes service

    Note, these tests must be run in-order. The order is defined by
    their definition order in this file. Nose follows this order according
    to its documentation:

        http://nose.readthedocs.org/en/latest/writing_tests.html
    """

    integration = True

    region = os.getenv('ECS_INTEGRATION_REGION', 'us-east-1')
    cluster = os.getenv('ECS_INTEGRATION_CLUSTER', 'default')

    def _successful_task(self):
        name = 'kingpin-integration-task-successful'
        return name, self._make_task(name, 'exit 0')

    def _slow_successful_task(self):
        name = 'kingpin-integration-task-slow-successful'
        return name, self._make_task(name, 'sleep 5; exit 0')

    def _failing_task(self):
        name = 'kingpin-integration-task-failing'
        return name, self._make_task(name, 'exit 1')

    def _slow_failing_task(self):
        name = 'kingpin-integration-task-slow-failing'
        return name, self._make_task(name, 'sleep 5; exit 1')

    def _hanging_task(self):
        name = 'kingpin-integration-task-hang'
        return name, self._make_task(name, 'tail -f /dev/null')

    def _make_task(self, name, command):
        return json.dumps({
            'family': name,
            'containerDefinitions': [{
                'name': name,
                'image': 'ubuntu:12.04',
                'cpu': 64,
                'memory': 64,
                'essential': True,
                'command': ['/bin/bash', '-c', command]
            }]})

    @contextmanager
    def _tmp_file(self, data):
        temp = tempfile.NamedTemporaryFile(suffix='.json', delete=False)
        temp.write(data)
        temp.close()
        yield temp.name
        os.unlink(temp.name)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=120)
    def integration_01a_run_successful_task(self):
        name, task = self._successful_task()
        with self._tmp_file(task) as definition:
            actor = ecs.RunTask(
                'Run %s' % name,
                {'region': self.region,
                 'cluster': self.cluster,
                 'task_definition': definition})
        yield actor.execute()

    @attr('aws', 'integration')
    @testing.gen_test(timeout=120)
    def integration_01b_run_slow_successful_task(self):
        name, task = self._slow_successful_task()
        with self._tmp_file(task) as definition:
            actor = ecs.RunTask(
                'Run %s' % name,
                {'region': self.region,
                 'cluster': self.cluster,
                 'task_definition': definition})
        yield actor.execute()

    @attr('aws', 'integration')
    @testing.gen_test(timeout=120)
    def integration_01c_run_failing_task(self):
        name, task = self._failing_task()
        with self._tmp_file(task) as definition:
            actor = ecs.RunTask(
                'Run %s' % name,
                {'region': self.region,
                 'cluster': self.cluster,
                 'task_definition': definition})
        with self.assertRaises(ecs.ECSTaskFailedException):
            yield actor.execute()

    @attr('aws', 'integration')
    @testing.gen_test(timeout=120)
    def integration_01d_run_slow_failing_task(self):
        name, task = self._slow_failing_task()
        with self._tmp_file(task) as definition:
            actor = ecs.RunTask(
                'Run %s' % name,
                {'region': self.region,
                 'cluster': self.cluster,
                 'task_definition': definition})
        with self.assertRaises(ecs.ECSTaskFailedException):
            yield actor.execute()

    @attr('aws', 'integration')
    @testing.gen_test(timeout=120)
    def integration_02a_run_slow_failing_task_dont_wait(self):
        name, task = self._slow_failing_task()
        with self._tmp_file(task) as definition:
            actor = ecs.RunTask(
                'Run %s' % name,
                {'region': self.region,
                 'cluster': self.cluster,
                 'task_definition': definition,
                 'wait': False})
        yield actor.execute()

    @attr('aws', 'integration')
    @testing.gen_test(timeout=120)
    def integration_03a_start_service_no_count(self):
        name, task = self._hanging_task()
        with self._tmp_file(task) as definition:
            actor = ecs.Service(
                'Starting service %s with count 0' % name,
                {'region': self.region,
                 'cluster': self.cluster,
                 'task_definition': definition,
                 'service_name': '%s_%s' % (name, '03'),
                 'count': 0})
        yield actor.execute()

    @attr('aws', 'integration')
    @testing.gen_test(timeout=120)
    def integration_03b_scale_service(self):
        name, task = self._hanging_task()
        with self._tmp_file(task) as definition:
            actor = ecs.Service(
                'Scaling service %s to count 1' % name,
                {'region': self.region,
                 'cluster': self.cluster,
                 'task_definition': definition,
                 'service_name': '%s_%s' % (name, '03'),
                 'use_existing_count': False,
                 'count': 1})
        yield actor.execute()

    @attr('aws', 'integration')
    @testing.gen_test(timeout=120)
    def integration_03c_scale_service_dont_wait(self):
        name, task = self._hanging_task()
        with self._tmp_file(task) as definition:
            actor = ecs.Service(
                'Scaling service %s to count 2' % name,
                {'region': self.region,
                 'cluster': self.cluster,
                 'task_definition': definition,
                 'service_name': '%s_%s' % (name, '03'),
                 'use_existing_count': False,
                 'count': 2})
        yield actor.execute()

    @attr('aws', 'integration')
    @testing.gen_test(timeout=120)
    def integration_03d_delete_service(self):
        name, task = self._hanging_task()
        with self._tmp_file(task) as definition:
            actor = ecs.Service(
                'Deleting service %s' % name,
                {'region': self.region,
                 'cluster': self.cluster,
                 'task_definition': definition,
                 'service_name': '%s_%s' % (name, '03'),
                 'state': 'absent'})
        yield actor.execute()

    @attr('aws', 'integration')
    @testing.gen_test(timeout=120)
    def integration_04a_start_service_scaled(self):
        name, task = self._hanging_task()
        with self._tmp_file(task) as definition:
            actor = ecs.Service(
                'Starting service %s with count 1' % name,
                {'region': self.region,
                 'cluster': self.cluster,
                 'task_definition': definition,
                 'service_name': '%s_%s' % (name, '04'),
                 'count': 1})
        yield actor.execute()

    @attr('aws', 'integration')
    @testing.gen_test(timeout=120)
    def integration_04b_scale_service_down(self):
        name, task = self._hanging_task()
        with self._tmp_file(task) as definition:
            actor = ecs.Service(
                'Scaling service %s to count 0' % name,
                {'region': self.region,
                 'cluster': self.cluster,
                 'task_definition': definition,
                 'service_name': '%s_%s' % (name, '04'),
                 'use_existing_count': False,
                 'count': 0})
        yield actor.execute()

    @attr('aws', 'integration')
    @testing.gen_test(timeout=120)
    def integration_04c_delete_service(self):
        name, task = self._hanging_task()
        with self._tmp_file(task) as definition:
            actor = ecs.Service(
                'Deleting service %s' % name,
                {'region': self.region,
                 'cluster': self.cluster,
                 'task_definition': definition,
                 'service_name': '%s_%s' % (name, '04'),
                 'state': 'absent'})
        yield actor.execute()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.hipchat`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The Hipchat Actors allow you to send messages to a HipChat room at stages
during your job execution. The actor supports dry mode by validating that the
configured API Token has access to execute the methods, without actually
sending the messages.

**Required Environment Variables**

:HIPCHAT_TOKEN:
  HipChat API Token

:HIPCHAT_NAME:
  HipChat message from name
  (defaults to ``Kingpin``)
"""

import logging
import os

from tornado import gen
from tornado import httpclient

from kingpin import utils
from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


API_CONTENT_TYPE = 'application/json'
API_URL = 'https://api.hipchat.com/v1'
API_MESSAGE_PATH = '%s/rooms/message' % API_URL
API_TOPIC_PATH = '%s/rooms/topic' % API_URL

TOKEN = os.getenv('HIPCHAT_TOKEN', None)
NAME = os.getenv('HIPCHAT_NAME', 'Kingpin')


class HipchatBase(base.HTTPBaseActor):

    """Simple Hipchat Abstract Base Object"""

    def __init__(self, *args, **kwargs):
        """Check required environment variables."""
        super(HipchatBase, self).__init__(*args, **kwargs)

        if not TOKEN:
            raise exceptions.InvalidCredentials(
                'Missing the "HIPCHAT_TOKEN" environment variable.')

        self._token = TOKEN
        self._name = self._validate_from_name(NAME)

    def _validate_from_name(self, name):
        """Parses and validates the FROM name.

        The username must be between 1 and 15 characters. If its not,
        we return a partial name to ensure that the push still works.

        Args:
            name: (Str) Proposed Hipchat message 'from' name

        Returns:
            A potentially modified string name thats valid.
        """
        return name[:15]

    def _build_potential_args(self, potential_args):
        """Builds a full set of arguments to pass to Hipchat.

        Appends the authentication token and a few other bits to the
        arguments supplied.

        Args:
            potential_Args: A hash of potential arguments.

        Returns:
            A larger hash of arguments.
        """
        potential_args['auth_token'] = self._token
        potential_args['from'] = self._name

        # If we're in 'dry run' mode, add the auth_test parameter
        if self._dry:
            potential_args['auth_test'] = True

        return potential_args

    @gen.coroutine
    @utils.retry(excs=exceptions.RecoverableActorFailure, retries=3)
    def _fetch_wrapper(self, *args, **kwargs):
        """Wrap the superclass _fetch method to catch known Hipchat errors."""
        try:
            res = yield self._fetch(*args, **kwargs)
        except httpclient.HTTPError as e:
            # These are HTTPErrors that we know about, and can log specific
            # error messages for.

            self.log.critical(e)
            if e.code in (401, 403):
                # "The authentication you provided is invalid."
                raise exceptions.InvalidCredentials(
                    'The "HIPCHAT_NAME" or "HIPCHAT_TOKEN" supplied is '
                    'invalid. %s' % e)
            else:
                # We ran into a problem we can't handle. Also, keep in mind
                # that @utils.retry() was used, so this error happened several
                # times before getting here. Raise it.
                raise exceptions.RecoverableActorFailure(
                    'Unexpected error from Hipchat API: %s' % e)

        raise gen.Return(res)


class Message(HipchatBase):

    """Sends a message to a room in HipChat.

    **Options**

    :room:
      (str) The string-name (or ID) of the room to send a message to

    :message:
      (str) Message to send

    **Examples**

    .. code-block:: json

       { "actor": "hipchat.Message",
         "desc": "Send a message!",
         "options": {
           "room": "Operations",
           "message": "Beginning Deploy: v1.2"
         }
       }

    **Dry Mode**

    Fully supported -- does not actually send messages to a room, but validates
    that the API credentials would have access to send the message using the
    HipChat ``auth_test`` optional API argument.
    """

    all_options = {
        'room': (str, REQUIRED, 'Hipchat room name'),
        'message': (str, REQUIRED, 'Message to send')
    }

    desc = "Sending Message to {room}"

    @gen.coroutine
    def _post_message(self, room_id, message,
                      message_format='html', notify=0,
                      color='yellow'):
        """Posts a message to Hipchat.

        https://www.hipchat.com/docs/api/method/rooms/message

        Args:
            room_id: (Str/Int) Name or ID of the room to post to.
            message: (Str) Required. The message body. 10,000 characters max.
            message_format: (Str) 'html' or 'text'.
            notify: (0/1) Whether or not this message should trigger a
                    notification for people in the room.
            color: (Str): Background color for message. One of "yellow", "red",
                          "green", "purple", "gray", or "random".

        Raises:
            gen.Return(<Dictionary of the response from Hipchat>)
        """
        args = self._build_potential_args({
            'room_id': room_id,
            'message': message,
            'message_format': message_format,
            'notify': notify,
            'color': color,
            'format': 'json',
        })
        url = self._generate_escaped_url(API_MESSAGE_PATH, args)
        res = yield self._fetch_wrapper(url)
        raise gen.Return(res)

    @gen.coroutine
    def _execute(self):
        """Executes an actor and yields the results when its finished.

        raises: gen.Return()
        """
        self.log.info('Sending message "%s" to Hipchat room "%s"' %
                      (self.option('message'), self.option('room')))
        res = yield self._post_message(self.option('room'),
                                       self.option('message'))

        # If we get 'None' or 'False' back, the actor failed.
        if not res:
            raise exceptions.RecoverableActorFailure(
                'Failed to send message to HipChat: %s' % res)

        # If we got here, the result is supposed to include 'success' as a key
        # and inside that key we can dig for the actual message. If the
        # response code is 202, we know that we didn't actually execute the
        # message send, but just validated the API token against the API.
        if 'success' in res:
            if res['success']['code'] == 202:
                self.log.info('API Token Validated: %s' %
                              res['success']['message'])

        raise gen.Return()


class Topic(HipchatBase):

    """Sets a HipChat room topic.

    **Options**

    -  ``room`` - The string-name (or ID) of the room to set the topic of
    -  ``topic`` - String of the topic to send

    **Examples**

    .. code-block:: json

      { "actor": "hipchat.Topic",
        "desc": "set the room topic",
        "options": {
          "room": "Operations",
          "topic": "Latest Deployment: v1.2"
        }
      }

    **Dry Mode**

    Fully supported -- does not actually set a room topic, but validates
    that the API credentials would have access to set the topic of the room
    requested.
    """

    all_options = {
        'room': (str, REQUIRED, 'Hipchat room name'),
        'topic': (str, REQUIRED, 'Topic to set')
    }

    desc = "Setting Room {room} topic"

    @gen.coroutine
    def _set_topic(self, room_id, topic):
        """Posts a message to Hipchat.

        https://www.hipchat.com/docs/api/method/rooms/topic

        Args:
            room_id: (Str/Int) Name or ID of the room to post to.
            topic: (Str) Required. The topic string, 250 char max

        Raises:
            gen.Return(<Dictionary of the response from Hipchat>)
        """
        args = self._build_potential_args({
            'room_id': room_id,
            'topic': topic,
            'format': 'json',
        })
        url = self._generate_escaped_url(API_TOPIC_PATH, args)

        # Note, we set post='' here to make sure we send a POST message, even
        # though were passing all of our arguments on the actual request line.
        res = yield self._fetch_wrapper(url, post='')
        raise gen.Return(res)

    @gen.coroutine
    def _execute(self):
        """Executes an actor and yields the results when its finished.

        raises: gen.Return()
        """
        self.log.info('Setting room "%s" topic to: %s' %
                      (self.option('room'), self.option('topic')))
        res = yield self._set_topic(self.option('room'),
                                    self.option('topic'))

        # If we get 'None' or 'False' back, the actor failed.
        if not res:
            raise exceptions.RecoverableActorFailure(
                'Failed to set room topic: %s' % res)

        # If we got here, the result is supposed to include 'success' as a key
        # and inside that key we can dig for the actual message. If the
        # response code is 202, we know that we didn't actually execute the
        # message send, but just validated the API token against the API.
        if 'success' in res:
            if res['success']['code'] == 202:
                self.log.info('API Token Validated: %s' %
                              res['success']['message'])

        raise gen.Return()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.rightscale.deployment`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. _Deployments:
    http://reference.rightscale.com/api1.5/resources/ResourceDeployments.html
"""

import logging

from tornado import gen

from kingpin.actors import exceptions
from kingpin.actors.rightscale import base
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'


class DeploymentBaseActor(base.RightScaleBaseActor):

    """Abstract Deployment Actor that provides some utility methods."""

    @gen.coroutine
    def _find_deployment(self, name):

        dep = yield self._client.find_by_name_and_keys(
            collection=self._client._client.deployments,
            name=name)

        raise gen.Return(dep)


class Create(DeploymentBaseActor):

    """Creates a RightScale deployment.

    Options match the documentation in RightScale:
    http://reference.rightscale.com/api1.5/resources/ResourceDeployments.html

    **Options**

    :name:
      The name of the deployment to be created.

    :description:
       The description of the deployment to be created.
       (*optional*)

    :server_tag_scope:
       The routing scope for tags for servers in the deployment.
       Can be 'deployment' or 'account'
       (*optional*, default: deployment)
    """

    all_options = {
        'name': (str, REQUIRED, 'The name of the deployment to be created.'),
        'description': (
            str, '',
            'The description of the deployment to be created.'),
        'server_tag_scope': (
            str, '',
            'The routing scope for tags for servers in the deployment.')
    }

    def __init__(self, *args, **kwargs):
        """Validate the user-supplied parameters at instantiation time."""

        super(Create, self).__init__(*args, **kwargs)

        allowed_scopes = ('deployment', 'account', '')

        scope = self.option('server_tag_scope')
        if scope not in allowed_scopes:
            raise exceptions.InvalidOptions(
                'server_tag_scope "%s" is not one of: %s' % (scope,
                                                             allowed_scopes))

    @gen.coroutine
    def _execute(self):

        dep = yield self._find_deployment(self.option('name'))
        if dep:
            raise exceptions.InvalidOptions(
                'Deployment "%s" already exists.' % self.option('name'))

        params = {'name': self.option('name'),
                  'description': self.option('description')}

        if self.option('server_tag_scope'):
            params['server_tag_scope'] = self.option('server_tag_scope')

        params = self._generate_rightscale_params('deployment', params)

        if self._dry:
            self.log.info('Would create a deployment %s' % self.option('name'))
            self.log.debug('Deployment params: %s' % params)
            raise gen.Return()

        self.log.info('Creating deployment %s' % self.option('name'))

        yield self._client.create_resource(
            self._client._client.deployments, params)


class Destroy(DeploymentBaseActor):

    """Deletes a RightScale deployment.

    Options match the documentation in RightScale:
    http://reference.rightscale.com/api1.5/resources/ResourceDeployments.html

    **Options**

    :name:
      The name of the deployment to be deleted.
    """

    all_options = {
        'name': (str, REQUIRED, 'The name of the deployment to be deleted.'),
    }

    @gen.coroutine
    def _execute(self):

        dep = yield self._find_deployment(self.option('name'))
        if not dep:
            raise exceptions.InvalidOptions(
                'Deployment "%s" does not exist.' % self.option('name'))

        info = (yield self._client.show(dep.self)).soul

        if self._dry:
            self.log.info('Would delete deployment %s' % info['name'])
            raise gen.Return()

        self.log.info('Deleting deployment %s' % info['name'])
        yield self._client.destroy_resource(dep)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.packagecloud`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The packagecloud actor allows you to perform maintenance operations on
repositories hosted by packagecloud.io using their API:

https://packagecloud.io/docs/api

**Required Environment Variables**

:PACKAGECLOUD_ACCOUNT:
  packagecloud account name, i.e. https://packagecloud.io/PACKAGECLOUD_ACCOUNT

:PACKAGECLOUD_TOKEN:
  packagecloud API Token
"""

import datetime
import logging
import os
import re
import sys

from tornado import gen

from tornado_rest_client import api

from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Charles McLaughlin <charles@nextdoor.com>'

ACCOUNT = os.getenv('PACKAGECLOUD_ACCOUNT', None)
TOKEN = os.getenv('PACKAGECLOUD_TOKEN', None)


class PackagecloudAPI(api.RestConsumer):

    ENDPOINT = 'https://packagecloud.io/api/v1/'
    CONFIG = {
        'attrs': {
            'packages': {
                'path': ('repos/%account%/%repo%/packages.json'
                         '?per_page={}'.format(sys.maxsize)),
                'http_methods': {'get': {}}
            },
            'delete': {
                'path': 'repos/%account%/%repo%/%distro_version%/%filename%',
                'http_methods': {'delete': {}}
            },
        },
        'auth': {
            'user': TOKEN,
            'pass': ''
        }
    }


class PackagecloudBase(base.BaseActor):

    """Simple packagecloud Abstract Base Object"""

    def __init__(self, *args, **kwargs):
        """Check required environment variables."""
        super(PackagecloudBase, self).__init__(*args, **kwargs)

        if not ACCOUNT:
            raise exceptions.InvalidCredentials(
                'Missing the "PACKAGECLOUD_ACCOUNT" environment variable.')

        if not TOKEN:
            raise exceptions.InvalidCredentials(
                'Missing the "PACKAGECLOUD_TOKEN" environment variable.')

        rest_client = api.RestClient(timeout=120)
        self._packagecloud_client = PackagecloudAPI(client=rest_client)

    @gen.coroutine
    def _get_all_packages(self, repo):
        """Simple method for fetching a dictionary of all packages in a repo

        Args:
            repo: name of the packagecloud repo to fetch from

        Returns:
            A hash of the packages.
        """
        packages = yield self._packagecloud_client.packages(
            token=TOKEN, account=ACCOUNT, repo=repo).http_get()
        raise gen.Return(packages)

    def _get_package_versions(self, name, packages):
        """Find all versions of a given package.

        Args:
            name: name of the package to look for
            packages: hash of all the packages, as returned by the API

        Returns:
            A hash of package versions sorted by creation date
        """
        versions = [{
            'created_at': datetime.datetime.strptime(
                package['created_at'], '%Y-%m-%dT%H:%M:%S.%fZ'),
            'distro_version': package['distro_version'],
            'filename': package['package_html_url'].split('/')[-1],
            'name': package['name']
        } for package in packages if package['name'] == name]

        versions.sort(key=lambda x: x.get('created_at'), reverse=False)
        return versions

    def _filter_packages(self, regex, packages):
        """Extracts a list of unique package names to delete

        Args:
            regex: regex of package names to delete
            packages: hash of all the packages, as returned by the API

        Returns:
            A list of unique package names that match the delete pattern.
            """
        pattern = re.compile(regex)
        packages_list_to_delete = {package['name'] for package in packages
                                   if pattern.match(package['name'])}

        self.log.debug('List of packages matching regex (%s): %s' %
                       (regex, packages_list_to_delete))

        return packages_list_to_delete

    @gen.coroutine
    def _delete(self, regex, repo, older_than=0,
                number_to_keep=0):
        """Generic packagecloud delete method, optionally supporting deleting
        old packages by date and/or keeping a certain number of packages.

        Args:
            regex: Regex of packages to delete, e.g. pkg1|pkg2
            repo: name of the packagecloud repo to delete from
            older_than: Delete packages created before this number of seconds
            number_to_keep: Keep at least this number of each package

        Returns:
            A list of the packages that were deleted
        """
        all_packages = yield self._get_all_packages(repo=repo)
        packages_list_to_delete = self._filter_packages(regex, all_packages)
        all_packages_deleted = []

        # Loop through each unique package to delete
        for name in packages_list_to_delete:
            package_versions = self._get_package_versions(
                name, all_packages)

            # Create a tally of the packages we delete -- usd to give the user
            # a final helpful log statement about the work we did.
            packages_deleted = []

            # Get a total count of the number of versions of this package in
            # the repo -- this variable will then be counted down as we loop,
            # to prevent us from deleting more than 'number_to_keep' packages.
            number_in_repo = len(package_versions)
            self.log.debug('Scanning %s versions (%s)' %
                           (name, number_in_repo))

            for package in package_versions:
                # Safety check -- if there aren't more than the number_to_keep
                # in the repo, then don't bother continuing through the loop
                # for this package. Break out and move to the next name in
                # packages_list_to_delete.
                if number_in_repo <= number_to_keep:
                    self.log.debug(
                        '%s has only %s package versions left, skipping'
                        % (name, number_in_repo))
                    break

                # If older_than (time in seconds) was supplied, figure out how
                # old the package is. If the package_age (in seconds) is
                # younger than allowed_age (in seconds), then skip to the next
                # package version in the set.
                if older_than:
                    package_age = (datetime.datetime.now() -
                                   package['created_at'])
                    allowed_age = datetime.timedelta(seconds=older_than)
                    if package_age <= allowed_age:
                        self.log.debug(
                            '%s/%s is only %s old, skipping deletion' %
                            (package['distro_version'],
                             package['name'], package_age))
                        continue

                # Finally if we got here, then we have enough packages left in
                # the repo, AND (optionally) this package is older than our
                # cutoff age... so delete the package.
                msg = '%s/%s/%s' % (
                    repo, package['distro_version'], package['filename'])
                if self._dry:
                    self.log.info('Would have deleted %s' % msg)
                else:
                    self.log.info('Deleting %s' % msg)

                    yield self._packagecloud_client.delete(
                        token=TOKEN, account=ACCOUNT, repo=repo,
                        distro_version=package['distro_version'],
                        filename=package['filename']
                    ).http_delete()

                # Decrement list of packages to track how many are left
                number_in_repo = number_in_repo - 1

                # Track *every* package we delete in one big dict -- this is
                # used purely for the unit tests to validate which packages
                # were deleted.
                all_packages_deleted.append(package)

                # Track that this package was deleted -- used in the parent for
                # loop to give the user a final tally of the packages that
                # were kept, and that were deleted.
                packages_deleted.append(package)

            # Print out the packages that were not deleted and left in the repo
            all_files = ['%s/%s' %
                         (package['distro_version'], package['filename'])
                         for package in package_versions]
            deleted_files = ['%s/%s' %
                             (package['distro_version'], package['filename'])
                             for package in packages_deleted]
            files_left = list(set(all_files) - set(deleted_files))
            self.log.debug('%s remaining packages: %s' % (name, files_left))

        raise gen.Return(all_packages_deleted)


class Delete(PackagecloudBase):

    """Deletes packages from a PackageCloud repo.

    Searches for packages that match the `packages_to_delete` regex pattern and
    deletes them.  If `number_to_keep` is set, we always at least this number
    of versions of the given package intact in the repo.  Also if
    `number_to_keep` is set, the older versions of a package (based on upload
    time) packages will be deleted first effectively leaving newer packages
    in the repo.

    **Options**

    :number_to_keep:
      Keep at least this number of each package
      (defaults to *0*)

    :packages_to_delete:
      Regex of packages to delete, e.g. pkg1|pkg2

    :repo:
      Which packagecloud repo to delete from

    **Examples**

    .. code-block:: json

      { "desc": "packagecloud Delete example",
        "actor": "packagecloud.Delete",
        "options": {
          "number_to_keep": 10,
          "packages_to_delete": "deleteme",
          "repo": "test"
        }
      }

    """

    all_options = {
        'number_to_keep': (
            int, 0,
            'Keep at least this number of each package'),
        'packages_to_delete': (
            str, REQUIRED,
            'Regex of packages to delete, e.g. pkg1|pkg2'),
        'repo': (
            str, REQUIRED,
            'Which packagecloud repo to delete from'),
    }

    desc = "Deleting {repo}/{packages_to_delete} (keeping {number_to_keep})"

    def __init__(self, *args, **kwargs):
        """Check required environment variables."""
        super(Delete, self).__init__(*args, **kwargs)

        try:
            re.compile(self.option('packages_to_delete'))
        except re.error:
            raise exceptions.InvalidOptions(
                'packages_to_delete is an invalid regex')

    @gen.coroutine
    def _execute(self):
        """Deletes all packages that match the `packages_to_delete` pattern"""
        yield self._delete(
            regex=self.option('packages_to_delete'),
            number_to_keep=self.option('number_to_keep'),
            repo=self.option('repo'))


class DeleteByDate(PackagecloudBase):

    """Deletes packages from a PackageCloud repo older than X.

    Adds additional functionality to the `Delete` class with a `older_than`
    option.  Only packages older than that number of seconds will be deleted.

    **Options**

    :number_to_keep:
      Keep at least this number of each package
      (defaults to *0*)

    :older_than:
      Delete packages created before this number of seconds

    :packages_to_delete:
      Regex of packages to delete, e.g. pkg1|pkg2

    :repo:
      Which packagecloud repo to delete from

    **Examples**

    .. code-block:: json

      { "desc": "packagecloud DeleteByDate example",
        "actor": "packagecloud.DeleteByDate",
        "options": {
          "number_to_keep": 10,
          "older_than": 600,
          "packages_to_delete": "deleteme",
          "repo": "test"
        }
      }

    """

    all_options = {
        'number_to_keep': (
            int, 0,
            'Keep at least this number of each package'),
        'older_than': (
            int, REQUIRED,
            'Delete packages created before this number of seconds'),
        'packages_to_delete': (
            str, REQUIRED,
            'Regex of packages to delete, e.g. pkg1|pkg2'),
        'repo': (str, REQUIRED,
                 'Which packagecloud repo to delete from')
    }

    desc = "Deleting {repo}/{packages_to_delete} older than {older_than}"

    @gen.coroutine
    def _execute(self):
        yield self._delete(
            regex=self.option('packages_to_delete'),
            number_to_keep=self.option('number_to_keep'),
            older_than=self.option('older_than'),
            repo=self.option('repo'))


class WaitForPackage(PackagecloudBase):

    """Searches for a package that matches `name` and `version` until found or
    a timeout occurs.

    **Options**

    :name:
      Name of the package to search for as a regex

    :version:
      Version of the package to search for as a regex

    :repo:
      Which packagecloud repo to delete from

    :sleep:
      Number of seconds to sleep for between each search

    **Examples**

    .. code-block:: json

      { "desc": "packagecloud WaitForPackage example",
        "actor": "packagecloud.WaitForPackage",
        "options": {
          "name": "findme",
          "version": "0.1",
          "repo": "test",
          "sleep": 10,
        }
      }

    """

    all_options = {
        'name': (
            str, REQUIRED, 'Name of the package to search for as a regex'),
        'version': (
            str, '.*', 'Version of the package to search for as a regex'),
        'repo': (
            str, REQUIRED, 'Which packagecloud repo to search'),
        'sleep': (
            int, 10, 'Number of seconds to sleep for between each search')
    }

    desc = "Waiting for {repo}/{name}@{version} (up to {sleep}s)"

    def __init__(self, *args, **kwargs):
        """Check required environment variables."""
        super(WaitForPackage, self).__init__(*args, **kwargs)

        try:
            re.compile(self.option('name'))
        except re.error:
            raise exceptions.InvalidOptions(
                'name is an invalid regex')

        try:
            re.compile(self.option('version'))
        except re.error:
            raise exceptions.InvalidOptions(
                'version is an invalid regex')

    @gen.coroutine
    def _search(self, repo, name, version):
        """Searches for a given package until found or a timeout occurs.

        Args:
            repo: name of the repo to search
            name: Name of the package to search for as a regex
            version: Version of the package to search for as a regex

        Returns:
            A list of the packages that were found
        """

        all_packages = yield self._get_all_packages(repo=repo)
        self.log.debug('Found all packages: %s' % all_packages)

        name_pattern = re.compile(name)
        version_pattern = re.compile(version)

        matched_packages = [p for p in all_packages
                            if name_pattern.match(p['name']) and
                            version_pattern.match(p['version'])]

        raise gen.Return(matched_packages)

    @gen.coroutine
    def _execute(self):
        """Execute method for the WaitForPackage actor"""
        while True:
            self.log.info('Searching for %s %s...' %
                          (self.option('name'), self.option('version')))

            matched_packages = yield self._search(
                repo=self.option('repo'),
                name=self.option('name'),
                version=self.option('version'))

            if len(matched_packages) > 0:
                self.log.info('Found it!')
                raise gen.Return()

            self.log.debug('Not found, sleeping for (%s)'
                           % self.option('sleep'))
            yield gen.sleep(self.option('sleep'))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.rightscale.server_template`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. _ServerTemplates_:
    http://reference.rightscale.com/api1.5/resources/ResourceServerTemplates.html
    http://reference.rightscale.com/api1.5/resources/ResourceServerTemplateMultiCloudImages.html
"""

import logging
import mock

from tornado import gen
import requests

from kingpin.actors import exceptions
from kingpin.actors.utils import dry
from kingpin.actors.rightscale import base
from kingpin.actors.rightscale import alerts
from kingpin.constants import SchemaCompareBase
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


class ServerTemplateMultiCloudImages(SchemaCompareBase):

    """Provides JSON-Schema based validation of the supplied Server Template.

    Each image is a dictionary that must have an MCI image name, and optionally
    whether or not its the default image for this template.

    _Note, only one MCI may have the is_default flag!_

    .. code-block:: json

        [
          {
              "mci": "Ubuntu 14.04 HVM".
              "rev": 5,
              "is_default": true,
          },
          {
              "mci": "Ubuntu 14.04 EBS".
          },
        ]
    }
    """

    SCHEMA = {
        'type': ['array', 'null'],
        'uniqueItems': True,
        'items': {
            'type': 'object',
            'required': ['mci'],
            'additionalProperties': False,
            'properties': {
                'mci': {
                    'type': 'string',
                    'minLength': 1,
                },
                'rev': {
                    'type': ['integer', 'string'],
                },
                'is_default': {
                    'type': 'boolean',
                },
            }
        }
    }


class ServerTemplateRunnableBindings(SchemaCompareBase):

    """Provides JSON-Schema based validation of the supplied Runnable Bindings.

    Each runnable binding must have a recipe or right_script,
    and a rev. Note, you cannot have both a recipe AND right_script.

    .. code-block:: json

        [
          {
            "position": 1,
            "recipe": "some cute recipe name",
            "rev": 2
          }
        ]
    """

    SCRIPT = {
        'type': 'object',
        'required': ['right_script', 'rev'],
        'additionalProperties': False,
        'properties': {
            'right_script': {
                'type': 'string',
            },
            'rev': {
                'type': 'number',
            },
        }
    }

    RECIPE = {
        'type': 'object',
        'required': ['recipe'],
        'additionalProperties': False,
        'properties': {
            'recipe': {
                'type': 'string',
            },
            'rev': {
                'type': 'number',
            },
        }
    }

    SCHEMA = {
        'type': ['array', 'null'],
        'uniqueItems': True,
        'items': {
            'anyOf': [
                SCRIPT

                # NOTE: This is not complete
                # RECIPE,
            ]
        }
    }


class ServerTemplate(base.EnsurableRightScaleBaseActor):

    """Manages the state of an RightScale ServerTemplate.

    This actor is able to create, destroy, modify and commit ServerTemplate
    revisions.

    Options match the documentation in RightScale:
    http://reference.rightscale.com/api1.5/resources/ResourceServerTemplates.html

    **Committing**

    If you wish to commit the template, set the `commit` option to the commit
    message you wish to use. If any change is made to the resource, Kingpin
    will commit a new revision with this message.

    **Options**

    :name:
      (str) The name of the template to be updated.

    :state:
      (str) Present or Absent. Default: "present"

    :commit:
      (str, None) If present, upon making changes to the HEAD version of the
      template, we will commit a new revision to RightScale. The provided
      string will be used as the commit message. Default: None

    :commit_head_dependencies:
      (bool) Commit all HEAD revisions (if any) of the associated MultiCloud
      Images, RightScripts and Chef repo sequences. Default: False

    :freeze_repositories:
      (bool) Freeze the repositories on commit. Default: False

    :tags:
      (list, str) A list of tags to add to the ServerTemplate.
      Default: None

    :description:
      (str) The description of the MCI image itself.
      Default: ""

    :images:
      A list of dicts that each describe a single cloud and the image in that
      cloud to launch. See below for details.

    :runnable_bindings:
      A list of runnable binding references.

    :alerts:
      A list of :py:mod:`kingpin.actors.rightscale.alerts.AlertSpecSchema`
      compatible dictionaries.

    **Image Definitions**

    Each cloud image definition is a dictionary that takes a few keys.

    :mci:
      (str) The name of the MultiCloudImage

    :rev:
      (str, int) The revision of the MultiCloudImage to use. Default: `0`

    :is_default:
      (bool) Whether or not this is the default MultiCloudImage for the
      template.

    **Examples**

    .. code-block:: json

        { "actor": "rightscale.ServerTemplate",
          "options": {
              "name": "Linux Server",
              "description": "this is a simple linux host",
              "images": [
                { "mci": "Ubuntu 14.04 HVM",
                  "rev": 5,
                  "is_default": true },
                { "mci": "Ubuntu 14.04 EBS" }
              ],
              "boot_bindings": [
                { "right_script": "Boot Script",
                  "rev": 2 }
              ]
          }
        }

    """

    all_options = {
        'name': (str, REQUIRED, 'The name of the template to be updated'),
        'commit': (str, None, 'Commit a new revision if changes made'),
        'commit_head_dependencies': (bool, False, (
            'Commit all HEAD revisions (if any) of the associated MultiCloud'
            'Images, RightScripts and Chef repo sequences.')
        ),
        'freeze_repositories': (bool, False,
                                'Freeze the repositories on commit'),
        'description': (str, None, 'The description of the to be updated'),
        'tags': ((list, str), None, 'List of tags to apply to the template'),
        'images': (ServerTemplateMultiCloudImages, None, (
            'A list of dicts that include our MultiCloudImage names'
            ', revisions (or default is HEAD), and optionally a '
            'is_default flag')),
        'boot_bindings': (ServerTemplateRunnableBindings, None,
                          'A list of scripts to run at boot time'),
        'operational_bindings': (ServerTemplateRunnableBindings, None, (
            'A list of scripts that can be run after'
            ' the host is operational')),
        'decommission_bindings': (ServerTemplateRunnableBindings, None, (
            'A list of scripts to run at decommission time')),
        'alerts': (alerts.AlertSpecsSchema, [], (
            'A list of dicts with AlertSpec parameters')),
    }

    unmanaged_options = [
        'commit', 'commit_head_dependencies', 'freeze_repositories', 'name'
    ]

    desc = 'ServerTemplate: {name}'

    def __init__(self, *args, **kwargs):
        """Validate the user-supplied parameters at instantiation time."""

        super(ServerTemplate, self).__init__(*args, **kwargs)
        self.changed = False

        self.st = None
        self.boot_bindings = []
        self.desired_boot_bindings = []
        self.operational_bindings = []
        self.desired_operational_bindings = []
        self.decommission_bindings = []
        self.desired_decommission_bindings = []
        self.tags = []
        self.images = {}
        self.alert_specs = None

        self.params = self._generate_rightscale_params(
            prefix='server_template',
            params={
                'description': self.option('description'),
                'name': self.option('name')
            })

        self._verify_one_default_image()

    def _verify_one_default_image(self):
        """Parses through the supplied images and finds the default one.

        If there are more than one default image defined, throws an exception.

        args:
            images: ServerTemplateMultiCloudImages object
        """
        default_image = [
            i['mci'] for i in self.option('images') if i.get('is_default')]

        if len(default_image) > 1:
            raise exceptions.InvalidOptions(
                'Only one image may have is_default set: %s' %
                ', '.join(default_image))

    @gen.coroutine
    def _precache(self):
        self.log.debug('Searching for ServerTemplate')
        st = yield self._client.find_by_name_and_keys(
            collection=self._client._client.server_templates,
            name=self.option('name'),
            revision=0)

        # Default searches return us an empty list if there are no matching
        # resources, or return us the exact resource we're looking for. Thus,
        # if the return value is not a list, then we know we got the
        # ServerTemplate back.
        if not isinstance(st, list):
            self.st = st

        # Build up our MCI "references" ... that is, take the plain text names
        # that the user submitted and turn them into proper HREFs and store
        # that as the desired "MCI References" state.
        #
        #  This dictionary ends up looking like this:
        #
        #  {
        #    "/api/multi_cloud_images/414637003": {
        #      "default": True,
        #    },
        #    "/api/multi_cloud_images/414123004": {
        #      "default": False,
        #    }
        #  }
        self.desired_images = {}
        tasks = []
        for image in self.option('images'):
            tasks.append(self._get_mci_href(image))
        yield tasks

        # Build up all of our runnable binding references
        self.desired_boot_bindings = yield self._generate_bindings(
            self.option('boot_bindings'), 'boot')
        self.desired_operational_bindings = yield self._generate_bindings(
            self.option('operational_bindings'), 'operational')
        self.desired_decommission_bindings = yield self._generate_bindings(
            self.option('decommission_bindings'), 'decommission')

        # If we got a list back, return None because that means that the
        # ServerTemplate doesn't already exist.
        if isinstance(st, list) and len(st) == 0:
            self.st = mock.MagicMock(name='<mocked template>')
            self.st.href = None
            self.st.soul = {
                'name': None,
                'description': None,
            }

            self.tags = []
            self.alert_specs = None
            self.images = {}
            raise gen.Return()

        # Get the boot, operational and decommission runnable bindings
        # associated with the ServerTemplate and store them.
        (self.boot_bindings, self.operational_bindings,
         self.decommission_bindings) = yield self._get_bindings()

        # Get the list of tags associated with the existing ST
        self.tags = (yield self._get_resource_tags(self.st))

        # Get the _live_ images associated with the server template
        # {
        #   "/api/multi_cloud_images/414637003": {
        #     "default": True,
        #     "map_href": "/api/server_template_multi_cloud_images/1234",
        #     "map_obj": <object itself for possible deletion>
        #   }
        # }
        self.images = yield self._get_mci_mappings()

        # Create a list of RightScale alert actors. Each one of these actors
        # will be responsible for ensuring that a specific AlertSpec is up to
        # date. However, none of these actors will handle deleting any
        # unexpected AlertSpecs.

        self.alert_specs = alerts.AlertSpecsBase(
            desc=('ServerTemplate AlertSpecs: %s' %
                  self.st.soul['name']),
            options={
                'href': self.st.href,
                'specs': self.option('alerts')
            },
            dry=self._dry,
        )
        self.alert_specs._client = self._client
        yield self.alert_specs._precache()

    @gen.coroutine
    def _get_mci_href(self, image):
        name = image['mci']
        revision = image.get('revision', 0)
        default = image.get('is_default', False)

        try:
            mci = yield self._client.find_by_name_and_keys(
                collection=self._client._client.multi_cloud_images,
                name=name, revision=revision)
        except StopIteration:
            return

        if not mci:
            raise exceptions.InvalidOptions(
                'Invalid MCI Name/Rev supplied: %s/%s' % (name, revision))

        self.desired_images[mci.href] = {
            'default': default,
        }
        self.log.debug('Discovered %s (%s) -> %s' % (name, revision, mci.href))

    @gen.coroutine
    def _get_mci_mappings(self):
        if not self.st.href:
            raise gen.Return()

        try:
            raw = yield self._client.find_by_name_and_keys(
                collection=self._client._client.server_template_multi_cloud_images,  # nopep8
                server_template_href=self.st.href)
        except StopIteration:
            return
        if not isinstance(raw, list):
            raw = [raw]

        images = {}
        for mci_map in raw:
            self.log.debug('Existing MCI Mapping -> %s (default: %s)' %
                           (mci_map.href, mci_map.soul['is_default']))
            images[mci_map.links['multi_cloud_image']] = {
                'default': mci_map.soul['is_default'],
                'map_href': mci_map.href,
                'map_obj': mci_map
            }

        raise gen.Return(images)

    @gen.coroutine
    def _get_state(self):
        if not self.st or self.st.href is None:
            raise gen.Return('absent')

        raise gen.Return('present')

    @gen.coroutine
    def _set_state(self):
        if self.option('state') == 'absent':
            yield self._delete_st()
        else:
            yield self._create_st()

    @gen.coroutine
    @dry('Would have created the ServerTemplate')
    def _create_st(self):
        self.log.info('Creating ServerTemplate')
        self.st = yield self._client.create_resource(
            self._client._client.server_templates, self.params)
        self.changed = True

    @gen.coroutine
    @dry('Would have deleted ServerTemplate')
    def _delete_st(self):
        self.log.info('Deleting ServerTemplate')
        yield self._client.destroy_resource(self.st)
        self.changed = True

    @gen.coroutine
    def _get_description(self):
        raise gen.Return(self.st.soul['description'])

    @gen.coroutine
    @dry('Would have updated the template description')
    def _set_description(self):
        desc = self.option('description')
        self.log.info('Updating description: %s' % desc)
        self.st = yield self._client.update(self.st, self.params)
        self.changed = True

    @gen.coroutine
    def _get_tags(self):
        raise gen.Return(self.tags)

    @gen.coroutine
    def _set_tags(self):
        existing_tags = yield self._get_tags()
        new_tags = self.option('tags')

        # What tags should we add, delete?
        to_add = list(set(new_tags) - set(existing_tags))
        to_delete = list(set(existing_tags) - set(new_tags))

        if to_add:
            yield self._add_resource_tags(resource=self.st, tags=to_add)
            self.changed = True
        if to_delete:
            yield self._delete_resource_tags(resource=self.st, tags=to_delete)
            self.changed = True

    @gen.coroutine
    def _get_images(self):
        # Note, this method is never used.. just a placeholder. See
        # _compare_images() below instead.
        raise gen.Return()

    @gen.coroutine
    def _compare_images(self):
        # Copy the self.images dict, and then strip each key of the
        # map_href param. At this point, it should look identical to
        # our self.desired_images dict.
        existing_images = {}
        for image in list(self.images.keys()):
            existing_images[image] = {
                'default': self.images[image]['default']
            }

        raise gen.Return(existing_images == self.desired_images)

    @gen.coroutine
    def _set_images(self):
        to_add = [href for href in list(self.desired_images.keys())
                  if href not in list(self.images.keys())]
        to_delete = [href for href in list(self.images.keys())
                     if href not in list(self.desired_images.keys())]

        tasks = []
        for href in to_add:
            tasks.append(self._create_mci_reference(href))
        yield tasks

        yield self._ensure_mci_default()

        tasks = []
        for href in to_delete:
            tasks.append(self._delete_mci_reference(
                self.images[href]['map_obj']))
        yield tasks

    @gen.coroutine
    def _set_operational_bindings(self):
        yield self._set_bindings(
            self.desired_operational_bindings,
            self.operational_bindings,
            'operational')

    @gen.coroutine
    def _compare_operational_bindings(self):
        raise gen.Return(self._compare_bindings(
            self.desired_operational_bindings,
            self.operational_bindings))

    @gen.coroutine
    def _get_operational_bindings(self):
        raise gen.Return(self.operational_bindings)

    @gen.coroutine
    def _set_boot_bindings(self):
        yield self._set_bindings(
            self.desired_boot_bindings,
            self.boot_bindings,
            'boot')

    @gen.coroutine
    def _compare_boot_bindings(self):
        raise gen.Return(self._compare_bindings(
            self.desired_boot_bindings,
            self.boot_bindings))

    @gen.coroutine
    def _get_boot_bindings(self):
        raise gen.Return(self.boot_bindings)

    @gen.coroutine
    def _set_decommission_bindings(self):
        yield self._set_bindings(
            self.desired_decommission_bindings,
            self.decommission_bindings,
            'decommission')

    @gen.coroutine
    def _compare_decommission_bindings(self):
        raise gen.Return(self._compare_bindings(
            self.desired_decommission_bindings,
            self.decommission_bindings))

    @gen.coroutine
    def _get_decommission_bindings(self):
        raise gen.Return(self.decommission_bindings)

    @gen.coroutine
    def _get_bindings(self):
        all_bindings = yield self._client.show(
            self.st.runnable_bindings)

        boot = [b for b in all_bindings if b.soul['sequence'] == 'boot']
        boot.sort(key=lambda x: x.soul['position'])
        operational = [b for b in all_bindings
                       if b.soul['sequence'] == 'operational']
        operational.sort(key=lambda x: x.soul['position'])
        decommission = [b for b in all_bindings
                        if b.soul['sequence'] == 'decommission']
        decommission.sort(key=lambda x: x.soul['position'])

        raise gen.Return((boot, operational, decommission))

    @gen.coroutine
    def _generate_bindings(self, bindings, sequence):
        new_bindings = []

        if not bindings:
            raise gen.Return(new_bindings)

        position = 0
        for config in bindings:
            position += 1

            self.log.debug('Searching for %s (rev: %s)' %
                           (config['right_script'], config.get('rev')))

            try:
                raw = yield self._client.find_by_name_and_keys(
                    collection=self._client._client.right_scripts,
                    exact=True,
                    name=config['right_script'])
            except StopIteration:
                continue

            # If we got nothing back (empty list, or None), throw an exception
            if raw is None or not raw:
                raise exceptions.InvalidOptions(
                    'Ubable to find RightScript: %s' % config['right_script'])

            # If only one item is returned, and the user didn't explicitly set
            # the revision number, then we pick one for them. If they _did_
            # explicitly set the revision number, we'll fail gracefully.
            if not isinstance(raw, list):
                if raw.soul['revision'] is not config.get('rev'):
                    raise exceptions.InvalidOptions(
                        'Invalid Binding Config: %s' % config)

                new_bindings.append({
                    'position': position,
                    'right_script_href': raw.href,
                    'sequence': sequence
                })
                continue
            else:
                # Ok, if we got a list back, we need to search through the list
                # and look for the desired revision. Again, if the user
                # specified one explicitly, then we'll look for that.
                # Otherwise, we pick the "latest" release (HEAD/0).
                try:
                    res = [res for res in raw
                           if res.soul['revision'] == config.get('rev')][0]
                    new_bindings.append({
                        'position': position,
                        'right_script_href': res.href,
                        'sequence': sequence
                    })
                    continue
                except IndexError:
                    raise exceptions.InvalidOptions(
                        'Invalid Binding Config: %s' % config)

        new_bindings.sort(key=lambda x: x['position'])
        raise gen.Return(new_bindings)

    def _compare_bindings(self, desired, existing):
        existing = [
            {'position': e.soul['position'],
             'right_script_href': e.links['right_script']}
            for e in existing]
        desired = [
            {'position': d['position'],
             'right_script_href': d['right_script_href']}
            for d in desired]

        return desired == existing

    @gen.coroutine
    @dry('Would have replaced the {2} runnable bindings')
    def _set_bindings(self, params_to_add, bindings_to_delete, name):
        tasks = []
        for binding in bindings_to_delete:
            self.log.info('Removing binding %s' % binding.href)
            tasks.append(self._client.destroy_resource(binding))
            self.changed = True
        yield tasks

        for binding in params_to_add:
            self.log.info('Adding binding %s' % binding['right_script_href'])
            try:
                yield self._client.create_resource(
                    self.st.runnable_bindings,
                    self._generate_rightscale_params(
                        prefix='runnable_binding',
                        params={
                            'right_script_href': binding['right_script_href'],
                            'sequence': binding['sequence']
                        }))
                self.changed = True
            except StopIteration:
                pass

    @gen.coroutine
    @dry('Would have added MCI Mapping -> {0}')
    def _create_mci_reference(self, href):
        definition = self._generate_rightscale_params(
            prefix='server_template_multi_cloud_image',
            params={
                'multi_cloud_image_href': href,
                'server_template_href': self.st.href,
            }
        )

        self.log.info('Adding MCI %s to ServerTemplate' % href)
        try:
            yield self._client.create_resource(
                self._client._client.server_template_multi_cloud_images,
                definition)
            self.changed = True
        except StopIteration:
            pass

    @gen.coroutine
    @dry('Would have deleted MCI reference {0.links[multi_cloud_image]}')
    def _delete_mci_reference(self, map_obj):
        self.log.info('Deleting MCI %s from ServerTemplate' %
                      map_obj.links['multi_cloud_image'])

        try:
            yield self._client.destroy_resource(map_obj)
        except requests.exceptions.HTTPError as e:
            if 'Default ServerTemplateMultiCloudImages' in e.response.text:
                raise exceptions.InvalidOptions(
                    'Cannot delete the current default MultiCloudImage '
                    'for this ServerTemplate. You must first re-assign a '
                    'new default image.')
        self.changed = True

    @gen.coroutine
    def _ensure_mci_default(self):
        """Sets the ServerTemplates Default MCI.

        After we've attached all of the MCIs to a ServerTemplate, we need to
        (optionally) set the "default" MCI to use if the user doesn't select
        one. This default can only be set after these attachments are made.
        Also, the default choice is then un-deletable from the MCI until a new
        default is selected.
        """
        # If there is no current default, then that means there are no MCIs
        # associated at all and we can't do anything. This only happens when
        # you're creating a new MCI from scratch.
        if 'default_multi_cloud_image' not in self.st.links:
            raise gen.Return()

        # Get the default MCI href as described by the user -- or just get the
        # first key in the list and treat that as the desired default.
        try:
            default_mci_href = [key for key in list(self.desired_images.keys())
                                if self.desired_images[key]['default'] is
                                True][0]
        except IndexError:
            default_mci_href = list(self.desired_images.keys())[0]

        # Compare the desired vs current default_multi_cloud_image_href. This
        # comparison is quick and doesn't require any API calls, so we do it
        # before we do anything else.
        current_default = self.st.links['default_multi_cloud_image']
        self.log.debug('Desired Default MCI HREF: %s' % default_mci_href)
        self.log.debug('Current Default MCI HREF: %s' % current_default)
        matching = (current_default == default_mci_href)

        # They match? Great, get out of here before we do anything bad!
        if matching:
            self.log.debug('Existing Default MCI HREF matches')
            raise gen.Return()

        # At this point, we know they don't match .. so we need to go off and
        # get all of the ServerTemplateMultiCloudImage references that link
        # these ServerTemplates to the MCIs.
        mci_refs = yield self._client.find_by_name_and_keys(
            collection=self._client._client.server_template_multi_cloud_images,
            server_template_href=self.st.href)

        # Final sanity check here -- if we're in DRY mode, get out!
        if self._dry:
            self.log.warning('Would have updated default MCI HREF to: %s'
                             % default_mci_href)
            raise gen.Return()

        # If we're not in DRY mode, AND we're updating the MCI reference to a
        # new default, then find the appropriate
        # server_template_muilti_cloud_image reference object. This should
        # always work, but some slowness in RightScales APIs has demonstrated
        # problems in the past, so we wrap this in a Try/Except block.
        try:
            new_default = [
                s for s in mci_refs
                if s.links['multi_cloud_image'] == default_mci_href][0]
        except IndexError:
            raise exceptions.InvalidOptions(
                'Unable to find the desired MultiCloud image '
                'attached to the ServerTemplate. This should never '
                'happen -- but can happen if the RightScale API is '
                'returning stale data. In that case, please wait and try '
                'again in the future.')

        # Finally, we have the desired new_default object... grab its URL
        # reference and lets make the API call.
        self.log.info('Making %s the default MCI' % default_mci_href)
        url = '%s/make_default' % new_default.links['self']
        yield self._client.make_generic_request(url, post=[])
        self.changed = True

    @gen.coroutine
    def _get_alerts(self):
        """Unnecessary method -- _compare_alerts is used instead"""
        raise gen.Return()

    @gen.coroutine
    def _compare_alerts(self):
        equals = True
        if self.alert_specs:
            equals = yield self.alert_specs._compare_specs()
        raise gen.Return(equals)

    @gen.coroutine
    def _set_alerts(self):
        yield self.alert_specs.execute()
        if self.alert_specs.changed:
            self.changed = True

    @gen.coroutine
    @dry('Would have committed HEAD to a revision')
    def _commit(self):
        self.log.info('Committing a new revision')
        params = {
            'commit_message': self.option('commit'),
            'freeze_repositories':
                str(self.option('freeze_repositories')).lower(),
            'commit_head_dependencies':
                str(self.option('commit_head_dependencies')).lower(),
        }

        ret = yield self._client.commit_resource(
            res=self.st,
            res_type=self._client._client.server_templates,
            params=params)

        self.log.info('Committed revision %s' % ret.soul['revision'])

    @gen.coroutine
    def _execute(self):
        yield super(ServerTemplate, self)._execute()

        # If we're deleting the MCI, then there is no need to continue after
        # we've done that.
        if self.option('state') == 'absent':
            raise gen.Return()

        # Finally, if we're committing and a change was made, commit!
        if self.changed and self.option('commit'):
            yield self._commit()



import logging

from boto.exception import NoAuthHandlerFound
from boto.exception import BotoServerError
from boto import utils
from tornado import testing
import botocore.exceptions
import mock

from kingpin.actors import exceptions
from kingpin.actors.aws import base
from kingpin.actors.aws import settings
import importlib

log = logging.getLogger(__name__)

# STATIC VALUES FOR TESTS
TARGET_GROUP_RESPONSE = {
    'ResponseMetadata': {
        'HTTPHeaders': {
            'content-length': '1228',
            'content-type': 'text/xml',
            'date': 'Tue, 13 Feb 2018 17:50:56 GMT',
            'x-amzn-requestid': '123-123-123-123-123'
        },
        'HTTPStatusCode': 200,
        'RequestId': '123-123-123-123-123',
        'RetryAttempts': 0
    },
    'TargetGroups': [{
        'HealthCheckIntervalSeconds': 30,
        'HealthCheckPath': '/',
        'HealthCheckPort': 'traffic-port',
        'HealthCheckProtocol': 'HTTP',
        'HealthCheckTimeoutSeconds': 5,
        'HealthyThresholdCount': 5,
        'LoadBalancerArns': [],
        'Matcher': {'HttpCode': '200'},
        'Port': 80,
        'Protocol': 'HTTP',
        'TargetGroupArn':
            'arn:aws:elb:us-east-1:123:targetgroup/unittest/123',
        'TargetGroupName': 'kingpin-integration-test',
        'UnhealthyThresholdCount': 2,
        'VpcId': 'vpc-123'
    }]
}


class TestBase(testing.AsyncTestCase):

    def setUp(self):
        super(TestBase, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        importlib.reload(base)

    @mock.patch('boto.iam.connection.IAMConnection')
    def test_missing_auth(self, mock_iam):
        mock_iam.side_effect = NoAuthHandlerFound('bad')
        with self.assertRaises(exceptions.InvalidCredentials):
            base.AWSBaseActor('Unit Test Action', {'region': 'fail'})

    def test_region_check(self):
        with self.assertRaises(exceptions.InvalidOptions):
            base.AWSBaseActor('Unit Test Action', {'region': 'fail'})

    def test_zone_check(self):
        actor = base.AWSBaseActor('Unit Test Action',
                                  {'region': 'us-west-1d'})
        self.assertEqual(actor.ec2_conn.region.name, 'us-west-1')

    @testing.gen_test
    def test_api_call_400(self):
        actor = base.AWSBaseActor('Unit Test Action', {})
        actor.elb_conn = mock.Mock()
        actor.elb_conn.get_all_load_balancers = mock.MagicMock()
        exc = BotoServerError(400, 'Bad Request')
        actor.elb_conn.get_all_load_balancers.side_effect = exc

        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor._find_elb('')

    @testing.gen_test
    def test_api_call_403(self):
        actor = base.AWSBaseActor('Unit Test Action', {})
        actor.elb_conn = mock.Mock()
        actor.elb_conn.get_all_load_balancers = mock.MagicMock()
        exc = BotoServerError(403, 'The security token')
        actor.elb_conn.get_all_load_balancers.side_effect = exc

        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor._find_elb('')

    @testing.gen_test
    def test_api_call_queue_400(self):
        actor = base.AWSBaseActor('Unit Test Action', {})
        actor.elb_conn = mock.Mock()
        actor.elb_conn.get_all_load_balancers = mock.MagicMock()
        exc = BotoServerError(400, 'Bad Request')
        actor.elb_conn.get_all_load_balancers.side_effect = exc

        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor.api_call_with_queueing(
                actor.elb_conn.get_all_load_balancers,
                queue_name='get_all_load_balancers')

    @testing.gen_test
    def test_api_call_queue_403(self):
        actor = base.AWSBaseActor('Unit Test Action', {})
        actor.elb_conn = mock.Mock()
        actor.elb_conn.get_all_load_balancers = mock.MagicMock()
        exc = BotoServerError(403, 'The security token')
        actor.elb_conn.get_all_load_balancers.side_effect = exc

        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor.api_call_with_queueing(
                actor.elb_conn.get_all_load_balancers,
                queue_name='get_all_load_balancers')

    @testing.gen_test
    def test_find_elb(self):
        actor = base.AWSBaseActor('Unit Test Action', {})
        actor.elb_conn = mock.Mock()
        actor.elb_conn.get_all_load_balancers.return_value = ['test']

        elb = yield actor._find_elb('')

        self.assertEqual(elb, 'test')
        self.assertEqual(actor.elb_conn.get_all_load_balancers.call_count, 1)

        actor.elb_conn.get_all_load_balancers.assert_called_with(
            load_balancer_names='')

    @testing.gen_test
    def test_find_elb_error(self):
        actor = base.AWSBaseActor('Unit Test Action', {})

        # Pretend the request worked, but there are no ELBs
        actor.elb_conn = mock.Mock()
        actor.elb_conn.get_all_load_balancers = mock.Mock(return_value=[])
        with self.assertRaises(base.ELBNotFound):
            yield actor._find_elb('')

    @testing.gen_test
    def test_find_elb_exception_error(self):
        actor = base.AWSBaseActor('Unit Test Action', {})

        # Pretend the request worked, but there are no ELBs
        actor.elb_conn = mock.Mock()
        actor.elb_conn.get_all_load_balancers = mock.MagicMock()
        actor.elb_conn.get_all_load_balancers.side_effect = BotoServerError(
            400, 'LoadBalancerNotFound')
        with self.assertRaises(base.ELBNotFound):
            yield actor._find_elb('')

        # Pretend the request worked, but there are no ELBs
        actor.elb_conn.get_all_load_balancers.side_effect = BotoServerError(
            401, 'SomeOtherError')
        with self.assertRaises(BotoServerError):
            yield actor._find_elb('')

    @testing.gen_test
    def test_find_target_group(self):
        actor = base.AWSBaseActor('Unit Test Action', {})
        c_mock = mock.Mock()
        c_mock.describe_target_groups.return_value = TARGET_GROUP_RESPONSE
        actor.elbv2_conn = c_mock

        target = yield actor._find_target_group('123')
        self.assertEqual(
            target,
            'arn:aws:elb:us-east-1:123:targetgroup/unittest/123')
        c_mock.describe_target_groups.assert_called_with(
            Names=['123'])

    @testing.gen_test
    def test_find_target_group_too_many_results(self):
        actor = base.AWSBaseActor('Unit Test Action', {})
        c_mock = mock.Mock()
        resp = TARGET_GROUP_RESPONSE.copy()
        resp['TargetGroups'] = [
            TARGET_GROUP_RESPONSE['TargetGroups'][0],
            TARGET_GROUP_RESPONSE['TargetGroups'][0],
            TARGET_GROUP_RESPONSE['TargetGroups'][0],
        ]
        c_mock.describe_target_groups.return_value = resp
        actor.elbv2_conn = c_mock

        with self.assertRaises(base.ELBNotFound):
            yield actor._find_target_group('123')

    @testing.gen_test
    def test_find_target_group_exception_error(self):
        actor = base.AWSBaseActor('Unit Test Action', {})
        c_mock = mock.Mock()
        exc = botocore.exceptions.ClientError({'Error': {'Code': ''}}, 'Test')
        c_mock.describe_target_groups.side_effect = exc
        actor.elbv2_conn = c_mock

        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            yield actor._find_target_group('123')

    @testing.gen_test
    def test_get_meta_data(self):
        actor = base.AWSBaseActor('Unit Test Action', {})

        with mock.patch.object(utils, 'get_instance_metadata') as md:
            md.return_value = {'ut-key': 'ut-value'}
            meta = yield actor._get_meta_data('ut-key')

        self.assertEqual(meta, 'ut-value')

    @testing.gen_test
    def test_get_meta_data_error(self):
        actor = base.AWSBaseActor('Unit Test Action', {})

        with mock.patch.object(utils, 'get_instance_metadata') as md:
            md.return_value = {}
            with self.assertRaises(base.InvalidMetaData):
                yield actor._get_meta_data('ut-key')

        with mock.patch.object(utils, 'get_instance_metadata') as md:
            md.return_value = {'key': 'value'}
            with self.assertRaises(base.InvalidMetaData):
                yield actor._get_meta_data('ut-key')

    @testing.gen_test
    def test_policy_doc_to_dict(self):
        policy_str = ''.join([
            '%7B%22Version%22%3A%20%222012-10-17%22%2C%20',
            '%22Statement%22%3A%20%5B%7B%22Action%22%3A%20%5B',
            '%22s3%3ACreate%2A%22%2C%20%22s3%3AGet%2A%22%2C%20',
            '%22s3%3APut%2A%22%2C%20%22s3%3AList%2A%22%5D%2C%20',
            '%22Resource%22%3A%20%5B',
            '%22arn%3Aaws%3As3%3A%3A%3Akingpin%2A%2F%2A%22%2C%20',
            '%22arn%3Aaws%3As3%3A%3A%3Akingpin%2A%22%5D%2C%20',
            '%22Effect%22%3A%20%22Allow%22%7D%5D%7D'])
        policy_dict = {
            'Version': '2012-10-17',
            'Statement': [
                {'Action': [
                    's3:Create*',
                    's3:Get*',
                    's3:Put*',
                    's3:List*'],
                 'Resource': [
                    'arn:aws:s3:::kingpin*/*',
                    'arn:aws:s3:::kingpin*'],
                 'Effect': 'Allow'}]}

        actor = base.AWSBaseActor('Unit Test Action', {})
        ret = actor._policy_doc_to_dict(policy_str)
        self.assertEqual(ret, policy_dict)

    @testing.gen_test
    def test_parse_policy_json(self):
        actor = base.AWSBaseActor('Unit Test Action', {})

        # Should work fine by default with good data
        ret = actor._parse_policy_json('examples/aws.iam.user/s3_example.json')
        self.assertEqual(ret['Version'], '2012-10-17')

        # If the file doesn't exist, raise an exception
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            actor._parse_policy_json('junk')

    @testing.gen_test
    def test_parse_policy_json_none(self):
        actor = base.AWSBaseActor('Unit Test Action', {})
        ret = actor._parse_policy_json(None)
        self.assertEqual(ret, None)



"""Tests for the actors.hipchat package"""

from nose.plugins.attrib import attr

from tornado import testing

from kingpin.actors import hipchat
from kingpin.actors import exceptions
import importlib


__author__ = 'Matt Wise <matt@nextdoor.com>'


class IntegrationHipchatMessage(testing.AsyncTestCase):

    """Simple high level integration tests agains the HipChat API.

    These tests actually hit the HipChat API and test that the code
    works, as well as validate that the API token is working properly.

    Require HIPCHAT_TOKEN environment variable to be set.
    """

    integration = True

    @attr('hipchat', 'integration')
    @testing.gen_test(timeout=60)
    def integration_test_init_without_environment_creds(self):
        message = 'Unit test message'
        room = 'Operations'

        # Un-set the token now and make sure the init fails
        hipchat.TOKEN = None
        with self.assertRaises(exceptions.InvalidCredentials):
            hipchat.Message(
                'Unit Test Action',
                {'message': message, 'room': room}, dry=True)

        # Reload the hipchat library to re-get the token
        importlib.reload(hipchat)

    @attr('hipchat', 'integration')
    @testing.gen_test(timeout=60)
    def integration_test_execute_with_invalid_creds(self):
        message = 'Unit test message'
        room = 'unit_room'
        actor = hipchat.Message(
            'Unit Test Action',
            {'message': message, 'room': room}, dry=True)

        # Valid response test
        actor._token = 'Invalid'
        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor.execute()

    @attr('hipchat', 'integration')
    @testing.gen_test(timeout=60)
    def integration_test_execute_real(self):
        message = 'Unit test message'
        room = 'Operations'
        actor = hipchat.Message(
            'Unit Test Action',
            {'message': message, 'room': room}, dry=True)
        res = yield actor.execute()
        self.assertEqual(res, None)


class IntegrationHipchatTopic(testing.AsyncTestCase):

    """Simple high level integration tests agains the HipChat API.

    These tests actually hit the HipChat API and test that the code
    works, as well as validate that the API token is working properly.

    Require HIPCHAT_TOKEN environment variable to be set.
    """

    integration = True

    @attr('hipchat', 'integration')
    @testing.gen_test(timeout=60)
    def integration_test_init_without_environment_creds(self):
        topic = 'Unit test topic'
        room = 'Operations'

        # Un-set the token now and make sure the init fails
        hipchat.TOKEN = None
        with self.assertRaises(exceptions.InvalidCredentials):
            hipchat.Topic(
                'Unit Test Action',
                {'topic': topic, 'room': room}, dry=True)

        # Reload the hipchat library to re-get the token
        importlib.reload(hipchat)

    @attr('hipchat', 'integration')
    @testing.gen_test(timeout=60)
    def integration_test_execute_with_invalid_creds(self):
        topic = 'Unit test topic'
        room = 'unit_room'
        actor = hipchat.Topic(
            'Unit Test Action',
            {'topic': topic, 'room': room}, dry=True)

        # Valid response test
        actor._token = 'Invalid'
        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor.execute()

    @attr('hipchat', 'integration')
    @testing.gen_test(timeout=60)
    def integration_test_execute_real(self):
        topic = 'Unit test topic'
        room = 'Operations'
        actor = hipchat.Topic(
            'Unit Test Action',
            {'topic': topic, 'room': room}, dry=True)
        res = yield actor.execute()
        self.assertEqual(res, None)



"""Tests for the misc actors"""

import os
import time

from nose.plugins.attrib import attr

from tornado import testing

from kingpin.actors import exceptions
from kingpin.actors import misc


__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'


class IntegrationGenericHTTP(testing.AsyncTestCase):

    integration = True

    @attr('http', 'integration')
    @testing.gen_test(timeout=60)
    def integration_get(self):
        actor = misc.GenericHTTP('Test', {'url': 'http://httpbin.org/get'})

        yield actor.execute()

    @attr('http', 'integration')
    @testing.gen_test(timeout=60)
    def integration_post(self):
        actor = misc.GenericHTTP('Test', {
            'url': 'http://httpbin.org/post',
            'data': {'foo': 'bar'}})

        yield actor.execute()

    @attr('http', 'integration')
    @testing.gen_test(timeout=60)
    def integration_auth(self):
        actor = misc.GenericHTTP('Test', {
            'url': 'http://httpbin.org/basic-auth/unit/test',
            'username': 'unit',
            'password': 'test'})

        yield actor.execute()

    @attr('http', 'integration')
    @testing.gen_test(timeout=60)
    def integration_auth_fail(self):
        actor = misc.GenericHTTP('Test', {
            'url': 'http://httpbin.org/basic-auth/unit/test',
            'username': 'wrong',
            'password': 'pass'})

        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor.execute()


class IntegrationMacro(testing.AsyncTestCase):

    integration = True

    @attr('http', 'integration')
    @testing.gen_test
    def integration_execute(self):
        actor = misc.Macro('Test', {
            'macro': 'examples/test/sleep.json',
            'tokens': dict(os.environ)})

        start = time.time()
        yield actor.execute()
        end = time.time()
        runtime = end - start
        self.assertTrue(runtime > 0.1)

    @attr('http', 'integration')
    @testing.gen_test
    def integration_fail_without_env(self):
        # Actor should fail if tokens aren't passed for env. variables.
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            misc.Macro('Test', {'macro': 'examples/test/sleep.json'},
                       init_tokens={})

    @attr('http', 'integration')
    @testing.gen_test
    def integration_execute_remote(self):
        gh_src = 'https://raw.githubusercontent.com/Nextdoor/kingpin/master'
        # Successful __init__ on this actor validates downloading and parsing.
        misc.Macro('Test', {
            'macro': gh_src + '/examples/test/sleep.json',
            'tokens': dict(os.environ)})

    @attr('http', 'integration', 'dry')
    @testing.gen_test
    def integration_execute_dry(self):
        actor = misc.Macro('Test', {
            'macro': 'examples/test/sleep.json',
            'tokens': dict(os.environ)},
            dry=True)

        start = time.time()
        yield actor.execute()
        end = time.time()
        runtime = end - start
        self.assertTrue(runtime < 0.1)



"""Tests for the pingdom actors"""

from nose.plugins.attrib import attr

from tornado import testing

from kingpin import utils
from kingpin.actors import pingdom


__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'


class IntegrationPingdom(testing.AsyncTestCase):

    integration = True

    check_name = 'kingpin-integration-test'

    @attr('pingdom', 'integration')
    @testing.gen_test(timeout=60)
    def integration_01a_test_pause(self):

        actor = pingdom.Pause('Pause check', {'name': self.check_name})

        yield actor.execute()

        yield utils.tornado_sleep(30)  # Let Pingdom cache settle

        check = yield actor._get_check()

        self.assertEqual(check['status'], 'paused')

    @attr('pingdom', 'integration')
    @testing.gen_test(timeout=60)
    def integration_02a_test_unpause(self):

        actor = pingdom.Unpause('Unpause check', {'name': self.check_name})

        yield actor.execute()

        yield utils.tornado_sleep(30)  # Let Pingdom cache settle

        check = yield actor._get_check()

        self.assertEqual(check['status'], 'up')



"""Tests for the actors.hipchat package"""

import json
import mock
import io

from tornado import gen
from tornado import httpclient
from tornado import testing

from kingpin.actors import hipchat
from kingpin.actors import exceptions


__author__ = 'Matt Wise <matt@nextdoor.com>'


class FakeHTTPClientClass(object):

    '''Fake HTTPClient object for testing'''

    response_value = None

    @gen.coroutine
    def fetch(self, *args, **kwargs):
        raise gen.Return(self.response_value)


class FakeExceptionRaisingHTTPClientClass(object):

    '''Fake HTTPClient object for testing'''

    response_value = None

    @gen.coroutine
    def fetch(self, *args, **kwargs):
        raise self.response_value


class TestHipchatBase(testing.AsyncTestCase):

    """Unit tests for the Hipchat Message actor."""

    def setUp(self, *args, **kwargs):
        # For most tests, mock out the TOKEN
        super(TestHipchatBase, self).setUp()
        hipchat.TOKEN = 'Unittest'

    def test_validate_from_name(self):
        message = 'Unit test message'
        room = 'Operations'
        actor = hipchat.Message(
            'Unit Test Action',
            {'message': message, 'room': room})

        # Regular run
        self.assertEqual('Foo Bar', actor._validate_from_name('Foo Bar'))

    def test_build_potential_args(self):
        potential_args = {
            'foo': 'bar',
            'baz': 'bat',
        }
        expected_args = dict(potential_args)
        expected_args['auth_token'] = hipchat.TOKEN
        expected_args['from'] = hipchat.NAME

        expected_dry_args = dict(expected_args)
        expected_dry_args['auth_test'] = True

        message = 'Unit test message'
        room = 'Operations'
        actor = hipchat.Message(
            'Unit Test Action',
            {'message': message, 'room': room})

        # Regular run
        args = actor._build_potential_args(potential_args)
        self.assertEqual(args, expected_args)

        # Now in dry mode
        actor._dry = True
        args = actor._build_potential_args(potential_args)
        self.assertEqual(args, expected_dry_args)

    @testing.gen_test
    def test_init_without_environment_creds(self):
        # Un-set the token now and make sure the init fails
        hipchat.TOKEN = None
        with self.assertRaises(exceptions.InvalidCredentials):
            hipchat.Message('Unit Test Action',
                            {'room': 'test', 'message': 'test'})

    @testing.gen_test
    def test_init_with_missing_options(self):
        with self.assertRaises(exceptions.InvalidOptions):
            hipchat.Message('Unit Test Action', {})


class TestHipchatMessage(testing.AsyncTestCase):

    """Unit tests for the Hipchat Message actor."""

    def setUp(self, *args, **kwargs):
        # For most tests, mock out the TOKEN
        super(TestHipchatMessage, self).setUp()
        hipchat.TOKEN = 'Unittest'

    @testing.gen_test
    def test_execute(self):
        message = 'Unit test message'
        room = 'unit_room'
        actor = hipchat.Message(
            'Unit Test Action',
            {'message': message, 'room': room})

        # Valid response test
        response_dict = {'status': 'sent'}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPResponse(
            httpclient.HTTPRequest('/'), code=200,
            buffer=io.StringIO(response_body))

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeHTTPClientClass()
            m.return_value.response_value = http_response
            res = yield actor._execute()
            self.assertEqual(res, None)

    @testing.gen_test
    def test_execute_dry_mode_response(self):
        message = 'Unit test message'
        room = 'unit_room'
        actor = hipchat.Message(
            'Unit Test Action',
            {'message': message, 'room': room})

        # Valid response test
        response_dict = {'success': {'code': 202, 'type': 'Accepted',
                         'message': 'It worked'}}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPResponse(
            httpclient.HTTPRequest('/'), code=202,
            buffer=io.StringIO(response_body))

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeHTTPClientClass()
            m.return_value.response_value = http_response
            res = yield actor._execute()
            self.assertEqual(res, None)

    @testing.gen_test
    def test_execute_with_401(self):
        message = 'Unit test message'
        room = 'unit_room'
        actor = hipchat.Message(
            'Unit Test Action',
            {'message': message, 'room': room})

        # Valid response test
        response_dict = {'error': {'code': 401, 'type': 'Unauthorized',
                         'message': 'Auth token not found'}}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPError(
            code=401, response=response_body)

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeExceptionRaisingHTTPClientClass()
            m.return_value.response_value = http_response

            with self.assertRaises(exceptions.InvalidCredentials):
                yield actor._execute()

    @testing.gen_test
    def test_execute_with_403(self):
        message = 'Unit test message'
        room = 'unit_room'
        actor = hipchat.Message(
            'Unit Test Action',
            {'message': message, 'room': room})

        # Valid response test
        response_dict = {'error': {'code': 403, 'type': 'Forbidden',
                         'message': 'Hit the rate limit'}}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPError(
            code=403, response=response_body)

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeExceptionRaisingHTTPClientClass()
            m.return_value.response_value = http_response

            with self.assertRaises(exceptions.InvalidCredentials):
                yield actor._execute()

    @testing.gen_test
    def test_execute_with_unknown_exception(self):
        message = 'Unit test message'
        room = 'unit_room'
        actor = hipchat.Message(
            'Unit Test Action',
            {'message': message, 'room': room})

        # Valid response test
        response_dict = {'error': {'code': 123, 'type': 'Unknown',
                         'message': 'Auth token not found'}}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPError(
            code=123, response=response_body)

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeExceptionRaisingHTTPClientClass()
            m.return_value.response_value = http_response

            with self.assertRaises(exceptions.RecoverableActorFailure):
                yield actor._execute()

    @testing.gen_test
    def test_execute_with_empty_response(self):
        message = 'Unit test message'
        room = 'unit_room'
        actor = hipchat.Message(
            'Unit Test Action',
            {'message': message, 'room': room})

        @gen.coroutine
        def fake_post_message(*args, **kwargs):
            raise gen.Return(None)
        actor._post_message = fake_post_message

        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield actor._execute()


class TestHipchatTopic(testing.AsyncTestCase):

    """Unit tests for the Hipchat Message actor."""

    def setUp(self, *args, **kwargs):
        # For most tests, mock out the TOKEN
        super(TestHipchatTopic, self).setUp()
        hipchat.TOKEN = 'Unittest'

    @testing.gen_test
    def test_execute(self):
        topic = 'Unit test topic'
        room = 'unit_room'
        actor = hipchat.Topic(
            'Unit Test Action',
            {'topic': topic, 'room': room})

        # Valid response test
        response_dict = {'status': 'sent'}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPResponse(
            httpclient.HTTPRequest('/'), code=200,
            buffer=io.StringIO(response_body))

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeHTTPClientClass()
            m.return_value.response_value = http_response
            res = yield actor._execute()
            self.assertEqual(res, None)

    @testing.gen_test
    def test_execute_dry_mode_response(self):
        topic = 'Unit test topic'
        room = 'unit_room'
        actor = hipchat.Topic(
            'Unit Test Action',
            {'topic': topic, 'room': room})

        # Valid response test
        response_dict = {'success': {'code': 202, 'type': 'Accepted',
                         'message': 'It worked'}}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPResponse(
            httpclient.HTTPRequest('/'), code=202,
            buffer=io.StringIO(response_body))

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeHTTPClientClass()
            m.return_value.response_value = http_response
            res = yield actor._execute()
            self.assertEqual(res, None)

    @testing.gen_test
    def test_execute_with_401(self):
        topic = 'Unit test topic'
        room = 'unit_room'
        actor = hipchat.Topic(
            'Unit Test Action',
            {'topic': topic, 'room': room})

        # Valid response test
        response_dict = {'error': {'code': 401, 'type': 'Unauthorized',
                         'message': 'Auth token not found'}}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPError(
            code=401, response=response_body)

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeExceptionRaisingHTTPClientClass()
            m.return_value.response_value = http_response

            with self.assertRaises(exceptions.InvalidCredentials):
                yield actor._execute()

    @testing.gen_test
    def test_execute_with_unknown_exception(self):
        topic = 'Unit test topic'
        room = 'unit_room'
        actor = hipchat.Topic(
            'Unit Test Action',
            {'topic': topic, 'room': room})

        # Valid response test
        response_dict = {'error': {'code': 123, 'type': 'Unknown',
                         'message': 'Auth token not found'}}
        response_body = json.dumps(response_dict)
        http_response = httpclient.HTTPError(
            code=123, response=response_body)

        with mock.patch.object(actor, '_get_http_client') as m:
            m.return_value = FakeExceptionRaisingHTTPClientClass()
            m.return_value.response_value = http_response

            with self.assertRaises(exceptions.RecoverableActorFailure):
                yield actor._execute()

    @testing.gen_test
    def test_execute_with_empty_response(self):
        topic = 'Unit test topic'
        room = 'unit_room'
        actor = hipchat.Topic(
            'Unit Test Action',
            {'topic': topic, 'room': room})

        @gen.coroutine
        def fake_set_topic(*args, **kwargs):
            raise gen.Return(None)
        actor._set_topic = fake_set_topic

        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield actor._execute()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.base`
^^^^^^^^^^^^^^^^^^^^^^^^^^

Base Actor object class

An Actor object is a class that executes a single logical action
on a resource as part of your deployment structure. For example, you
may have an Actor that launches a server array in RightScale, or you
may have one that sends an email.

Each Actor object should do one thing, and one thing only. Its responsible
for being able to execute the operation in both 'dry' and 'non-dry' modes.

The behavior for 'dry' mode can contain real API calls, but should not make
any live changes. It is up to the developer of the Actor to define what
'dry' mode looks like for that particular action.
"""

import inspect
import json
import logging
import os
import sys
import time

from tornado import gen
from tornado import httpclient
from tornado import httputil

from kingpin import utils
from kingpin.actors import exceptions
from kingpin.actors.utils import timer
from kingpin.constants import REQUIRED, STATE

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


# If super-debug logging is enabled, then we turn on the URLLIB3 HTTP
# request logging. This is extremely verbose and insecure, but useful
# for troubleshooting. URLLIB3 is used by several actors (aws, rightscale),
# so we do this setup here in the base actor class.
if os.getenv('URLLIB_DEBUG', None):
    utils.super_httplib_debug_logging()

# Allow the user to override the default_timeout for all actors by setting an
# environment variable
DEFAULT_TIMEOUT = os.getenv('DEFAULT_TIMEOUT', 3600)


class LogAdapter(logging.LoggerAdapter):

    """Simple Actor Logging Adapter.

    Provides a common logging format for actors that uses the actors
    description and dry parameter as a prefix to the supplied log message.
    """

    def process(self, msg, kwargs):
        return ('[%s%s] %s' % (self.extra['dry'], self.extra['desc'], msg),
                kwargs)


class BaseActor(object):

    """Abstract base class for Actor objects."""

    # {
    #     'option_name': (type, default, "Long description of the option"),
    # }
    #
    # If `default` is REQUIRED then the option requires user specified input
    #
    # Example:
    # {
    #    'room': (str, REQUIRED, 'Hipchat room to notify'),
    #    'from': (str, 'Kingpin', 'User that sends the message')
    # }
    all_options = {}

    # Default description format
    desc = "{actor}"

    # Set the default timeout for the gen.with_timeout() wrapper that we use to
    # monitor and control the length of execution of a single Actor.
    default_timeout = DEFAULT_TIMEOUT

    # Context separators. These define the left-and-right identifiers of a
    # 'contextual token' in the actor. By default this is { and }, so a
    # contextual token looks like '{KEY}'.
    left_context_separator = '{'
    right_context_separator = '}'

    # Ensure that at __init__ time, if the self._options dict is not completely
    # filled in properly (meaning there are no left-over {KEY}'s), we throw an
    # exception. This will change in the future when we have some concept of a
    # second 'global runtime context object'.
    strict_init_context = True

    # Controls whether to remove escape characters from tokens that have been
    # escaped.
    remove_escape_sequence = True

    def __init__(self, desc=None, options={}, dry=False, warn_on_failure=False,
                 condition=True, init_context={}, init_tokens={},
                 timeout=None):
        """Initializes the Actor.

        Args:
            desc: (Str) description of the action being executed.
            options: (Dict) Key/Value pairs that have the options
                     for this action. Values should be primitives.
            dry: (Bool) or not this Actor will actually make changes.
            warn_on_failure: (Bool) Whether this actor ignores its return
                             value and always succeeds (but warns).
            condition: (Bool) Whether to run this actor.
            init_context: (Dict) Key/Value pairs used at instantiation
                time to replace {KEY} strings in the actor definition.
                This is usually driven by the group.Sync/Async actors.
            init_tokens: (Dict) Key/Value pairs passed into the actor that can
            be used for token replacement. Typically this is os.environ() plus
            some custom tokens. Set generally by the misc.Macro actor.
            timeout: (Str/Int/Float) Timeout in seconds for the actor.
        """
        self._type = '%s.%s' % (self.__module__, self.__class__.__name__)
        self._options = options
        self._desc = desc
        self._dry = dry
        self._warn_on_failure = warn_on_failure
        self._condition = condition
        self._init_context = init_context
        self._init_tokens = init_tokens

        self._timeout = timeout
        if timeout is None:
            self._timeout = self.default_timeout

        # strict about this -- but in the future, when we have a
        # runtime_context object, we may loosen this restriction).
        self._fill_in_contexts(
            context=self._init_context,
            strict=self.strict_init_context,
            remove_escape_sequence=self.remove_escape_sequence)

        self._setup_log()
        self._setup_defaults()
        self._validate_options()  # Relies on _setup_log() above

        # Fill in any options with the supplied initialization context. Be
        self.log.debug('Initialized (warn_on_failure=%s, '
                       'strict_init_context=%s,'
                       'remove_escape_sequence=%s)' %
                       (warn_on_failure, self.strict_init_context,
                        self.remove_escape_sequence))

    def __repr__(self):
        """Returns a nice name/description of the actor.

        Either the user has supplied a custom desc parameter to the actor,
        giving it a useful description for them. On the other hand, if an actor
        defines a custom ActorClass.desc field, that field is interpreted by
        this method an any variables that can be swapped in dynamically are.

        For example, if misc.Sleep.desc is 'Sleeping {sleep}s', this method
        will fill in the value of the option 'sleep' into the string, and then
        use that for the representation of the object.
        """
        if self._desc:
            return self._desc

        return self.__class__.desc.format(actor=self._type, **self._options)

    def _setup_log(self):
        """Create a customized logging object based on the LogAdapter."""
        name = '%s.%s' % (self.__module__, self.__class__.__name__)
        logger = logging.getLogger(name)
        dry_str = 'DRY: ' if self._dry else ''

        self.log = LogAdapter(logger, {'desc': self, 'dry': dry_str})

    def _setup_defaults(self):
        """Populate options with defaults if they aren't set."""

        for option, definition in list(self.all_options.items()):
            if option not in self._options:
                default = definition[1]
                if default is not REQUIRED:
                    self._options.update({option: default})

    def _validate_options(self):
        """Validate that all the required options were passed in.

        Args:
            options: A dictionary of options.

        Raises:
            exceptions.InvalidOptions
        """

        # Loop through all_options, and find the required ones
        required = [opt_name
                    for (opt_name, definition) in
                    list(self.all_options.items())
                    if definition[1] is REQUIRED]

        self.log.debug('Checking for required options: %s' % required)
        option_errors = []
        option_warnings = []
        for opt in required:
            if opt not in self._options:
                description = self.all_options[opt][2]
                option_errors.append('Option "%s" is required: %s' % (
                                     opt, description))

        for opt, value in list(self._options.items()):
            if opt not in self.all_options:
                option_warnings.append('Option "%s" is not expected by %s.' % (
                    opt, self.__class__.__name__))
                continue

            expected_type = self.all_options[opt][0]

            # Unicode is not a `str` but it is a `basestring`
            # Cast the passed value explicitly as a string
            if isinstance(value, str):
                value = str(value)

            # If the expected_type has an attribute 'valid', then verify that
            # the option passed in is one of those valid options.
            if hasattr(expected_type, 'validate'):
                try:
                    expected_type.validate(value)
                    continue
                except exceptions.InvalidOptions as e:
                    option_errors.append(e)

            # If the option type is Bool, try to convert the strings True/False
            # into booleans. If this doesn't work, siletly move on and let the
            # failure get caught below.
            if expected_type is bool:
                try:
                    value = self.str2bool(value, strict=True)
                    self._options[opt] = value
                except exceptions.InvalidOptions as e:
                    self.log.warning(e)

            if not (value is None or isinstance(value, expected_type)):
                message = 'Option "%s" has to be %s and is %s.' % (
                    opt, expected_type, type(value))
                option_errors.append(message)

        for w in option_warnings:
            self.log.warning(w)

        if option_errors:
            for e in option_errors:
                self.log.critical(e)
            raise exceptions.InvalidOptions(
                'Found %s issue(s) with passed options.' % len(option_errors))

    def option(self, name):
        """Return the value for a given Actor option."""

        return self._options.get(name)

    def readfile(self, path):
        """Return file contents as a string.

        Raises:
            InvalidOptions if file is not found, or readable.
        """

        try:
            with open(path) as f:
                contents = f.read()
        except IOError as e:
            raise exceptions.InvalidOptions(e)

        return contents

    @gen.coroutine
    def timeout(self, f, *args, **kwargs):
        """Wraps a Coroutine method in a timeout.

        Used to wrap the self.execute() method in a timeout that will raise an
        ActorTimedOut exception if an actor takes too long to execute.

        *Note, Tornado 4+ does not allow you to actually kill a task on the
        IOLoop.*  This means that all we are doing here is notifying the caller
        (through the raised exception) that a problem has happened.

        Fairly simple Actors should actually 'stop executing' when this
        exception is raised. Complex actors with very unique behaviors though
        (like the rightsacle.server_array.Execute actor) have the ability to
        continue to execute in the background until the Kingpin application
        quits. It is not the job of this method to try to kill these actors,
        but just to let the user know that a failure has happened.
        """

        # Get our timeout setting, or fallback to the default
        self.log.debug('%s.%s() deadline: %s(s)' %
                       (self._type, f.__name__, self._timeout))

        # Get our Future object but don't yield on it yet, This starts the
        # execution, but allows us to wrap it below with the
        # 'gen.with_timeout' function.
        fut = f(*args, **kwargs)

        # If no timeout is set (none, or 0), then we just yield the Future and
        # return its results.
        if not self._timeout:
            ret = yield fut
            raise gen.Return(ret)

        # Generate a timestamp in the future at which point we will raise
        # an alarm if the actor is still executing
        deadline = time.time() + float(self._timeout)

        # Now we yield on the gen_with_timeout function
        try:
            ret = yield gen.with_timeout(
                deadline, fut, quiet_exceptions=(exceptions.ActorTimedOut))
        except gen.TimeoutError:
            msg = ('%s.%s() execution exceeded deadline: %ss' %
                   (self._type, f.__name__, self._timeout))
            self.log.error(msg)
            raise exceptions.ActorTimedOut(msg)

        raise gen.Return(ret)

    def str2bool(self, v, strict=False):
        """Returns a Boolean from a variety of inputs.

        args:
            value: String/Bool
            strict: Whether or not to _only_ convert the known words into
            booleans, or whether to allow "any" word to be considered True
            other than the known False words.

        returns:
            A boolean
        """
        false = ('no', 'false', 'f', '0')
        true = ('yes', 'true', 't', '1')

        string = str(v).lower()

        if strict:
            if string not in true and string not in false:
                raise exceptions.InvalidOptions(
                    'Expected [%s, %s] but got: %s' %
                    (true, false, string))

        return string not in false

    def _check_condition(self):
        """Check if specified condition allows this actor to run.

        Evaluate self._condition to figure out if this actor should run.
        The only exception to simply casting this variable to bool is if
        the value of self._condition is a string "False" or string "0".
        """

        check = self.str2bool(self._condition)
        self.log.debug('Condition %s evaluates to %s' % (
            self._condition, check))
        return check

    def _fill_in_contexts(self, context={}, strict=True,
                          remove_escape_sequence=True):
        """Parses self._options and updates it with the supplied context.

        Parses the objects self._options dict (by converting it into a JSON
        string, substituting, and then turning it back into a dict) and the
        self._desc string and replaces any {KEY}s with the valoues from the
        context dict that was supplied.

        Args:
            strict: bool whether or not to allow missing context keys to be
                    skipped over.

        Raises:
            exceptions.InvalidOptions
        """
        # Inject contexts into Description
        try:
            self._desc = utils.populate_with_tokens(
                str(self),
                context,
                self.left_context_separator,
                self.right_context_separator,
                strict=strict,
                remove_escape_sequence=remove_escape_sequence)
        except LookupError as e:
            msg = 'Context for description failed: %s' % e
            raise exceptions.InvalidOptions(msg)

        # Inject contexts into condition
        try:
            self._condition = utils.populate_with_tokens(
                str(self._condition),
                context,
                self.left_context_separator,
                self.right_context_separator,
                strict=strict,
                remove_escape_sequence=remove_escape_sequence)
        except LookupError as e:
            msg = 'Context for condition failed: %s' % e
            raise exceptions.InvalidOptions(msg)

        # Convert our self._options dict into a string for fast parsing
        options_string = json.dumps(self._options)

        # Generate a new string with the values parsed out. At this point, if
        # any value is un-matched, an exception is raised and execution fails.
        # This stops execution during a dry run, before any live changes are
        # made.
        try:
            new_options_string = utils.populate_with_tokens(
                options_string,
                context,
                self.left_context_separator,
                self.right_context_separator,
                strict=strict,
                escape_sequence='\\\\',
                remove_escape_sequence=remove_escape_sequence)
        except LookupError as e:
            msg = 'Context for options failed: %s' % e
            raise exceptions.InvalidOptions(msg)

        # Finally, convert the string back into a dict and store it.
        self._options = json.loads(new_options_string)

    def get_orgchart(self, parent=''):
        """Construct organizational chart describing this actor.

        Return a list of actors handled by this actor. Most actors will return
        a list of just one object. Grouping actors will return a list of all
        actors that are called.

        orgchart object:
          id: unique string identifying this actor's instance.
          class: kingpin class name
          desc: actor description
          parent_id: organizational relationship. Same as `id` above.
        """

        return [{
            'id': str(id(self)),
            'desc': self._desc,
            'class': self.__class__.__name__,
            # 'options': self._options,  # May include tokens & ENV vars
            'parent_id': parent,
        }]

    @gen.coroutine
    @timer
    def execute(self):
        """Executes an actor and yields the results when its finished.

        Calls an actors private _execute() method and either returns the result
        (through gen.Return) or handles any exceptions that are raised.

        RecoverableActorFailure exceptions are potentially swallowed up (and
        warned) if the self._warn_on_failure flag is set. Otherwise, they're
        logged and re-raised. All other ActorException exceptions are caught,
        logged and re-raised.

        We have a generic catch-all exception handling block as well, because
        third party Actor classes may or may not catch all appropriate
        exceptions. This block is mainly here to prevent the entire app from
        failing due to a poorly written Actor.

        Raises:
            gen.Return(result)
        """
        self.log.debug('Beginning')

        # Any exception thats raised by an actors _execute() method will
        # automatically cause actor failure and we return right away.
        result = None

        if not self._check_condition():
            self.log.warning('Skipping execution. Condition: %s' %
                             self._condition)
            raise gen.Return()

        try:
            result = yield self.timeout(self._execute)
        except exceptions.ActorException as e:
            # If exception is not RecoverableActorFailure
            # or if warn_on_failure is not set, then escalate.
            recover = isinstance(e, exceptions.RecoverableActorFailure)
            if not recover or not self._warn_on_failure:
                self.log.critical(e)
                raise

            # Otherwise - flag this failure as a warning, and continue
            self.log.warning(e)
            self.log.warning(
                'Continuing execution even though a failure was '
                'detected (warn_on_failure=%s)' % self._warn_on_failure)
        except Exception as e:
            # We don't like general exception catch clauses like this, but
            # because actors can be written by third parties and automatically
            # imported, its impossible for us to catch every exception
            # possible. This is a failsafe thats meant to throw a strong
            # warning.
            log.critical('Unexpected exception caught! '
                         'Please contact the author (%s) and provide them '
                         'with this stacktrace' %
                         sys.modules[self.__module__].__author__)
            self.log.exception(e)
            raise exceptions.ActorException(e)
        else:
            self.log.debug('Finished successfully, return value: %s' % result)

        # If we got here, we're exiting the actor cleanly and moving on.
        raise gen.Return(result)


class EnsurableBaseActor(BaseActor):

    """Base Class for Actors that "ensure" the state of a resource.

    Many of our actors have a goal of ensuring that a particular resource is in
    a given state. This leads to a ton of boiler plate code to "get" the state
    of something, "compare" that to the desired state, and then maybe "set" the
    state.

    This actor provides a framework allowing the user to simply write the
    getters and setters (and optionally compare), and lets the rest of the
    actor handle the order of operations.

    **Required Methods:**

      :`_set_state`: Creates or destroys the resource depending on
                      the 'state' parameter that was passed in.

                      *Note: The 'state' parameter is automatically added to
                      the options. You do not need to define it.*
      :`_get_state`: Gets the current state of the resource.
      :`_set_[option]`: A 'setter' for each option name passed in.
      :`_get_[option]`: A 'getter' for each option name passed in.

    **Optional Methods:**

      :`_precache`: Called before any setters/getters are triggered. Used
                       to optionally populate a cache of data to make the
                       getters faster. For example, if you can make one API
                       call to get all of the data about a resource, then
                       store that data locally for fast access.

      :`_compare_[option]`: Optionally you can write your own comparison
                               method if you're not doing a pure string
                               comparison between the source and destination.

    **Examples**

    .. code-block:: python

        class MyClass(base.EnsurableBaseActor):

            all_options = {
                'name': (str, REQUIRED, 'Name of thing'),
                'description': (str, None, 'Description of thing')
            }

            unmanaged_options = ['name']

            @gen.coroutine
            def _set_state(self):
                if self.option('state') == 'absent':
                    yield self.conn.delete_resource(
                        name=self.option('name'))
                else:
                    yield self.conn.create_resource(
                        name=self.option('name'),
                        desc=self.option('description'))

            @gen.coroutine
            def _set_description(self):
                yield self.conn.set_desc_of_resource(
                    name=self.option('name'),
                    desc=self.option('description'))

            @gen.coroutine
            def _get_description(self):
                yield self.conn.get_desc_of_resource(
                    name=self.option('name'))
    """

    # A list of option names that are _not_ automatically managed. These are
    # useful if you have special behaviors like 'commit' on change, or if you
    # have parameters that are unmutable ('name').
    unmanaged_options = []

    def __init__(self, *args, **kwargs):
        # The 'state' parameter is a given, so make sure its set,
        self.all_options['state'] = (
            STATE, 'present', 'Desired state: present or absent')

        # Now go ahead and validate all of the user inputs the normal way
        super(EnsurableBaseActor, self).__init__(*args, **kwargs)

        # Generate a list of options that will be ensured ...
        self._ensurable_options = list(self.all_options.keys())
        for option in self.unmanaged_options:
            self._ensurable_options.remove(option)

        # Finally, do a class validation... make sure that we have actual
        # getter/setter methods for each of the options. This populates dicts
        # that provide references to the actual methods for execution later.
        self._gather_methods()

    def _gather_methods(self):
        """Generates pointers to the Getter and Setter methods.

        Walks through the list of options in self.all_options and discovers the
        pointers to the getter/setter methods. If any are missing, throws an
        exception quickly.
        """
        self.setters = {}
        self.getters = {}
        self.comparers = {}
        for option in self._ensurable_options:
            setter = '_set_%s' % option
            getter = '_get_%s' % option
            comparer = '_compare_%s' % option

            if not self._is_method(getter) or not self._is_method(setter):
                raise exceptions.UnrecoverableActorFailure(
                    'Invalid Actor Code Detected in %s: '
                    'Unable to find required methods: %s, %s'
                    % (self.__class__.__name__, setter, getter))

            if not self._is_method(comparer):
                @gen.coroutine
                def _comparer(option=option):
                    existing = yield self.getters[option]()
                    new = self.option(option)
                    raise gen.Return(existing == new)
                setattr(self, comparer, _comparer)
                # self.log.debug('Creating dynamic method %s' % comparer)

            self.setters[option] = getattr(self, setter)
            self.getters[option] = getattr(self, getter)
            self.comparers[option] = getattr(self, comparer)

    def _is_method(self, name):
        return hasattr(self, name) and inspect.ismethod(getattr(self, name))

    @gen.coroutine
    def _precache(self):
        """Override this method to pre-cache data in your actor.

        This method can be overridden to go off and pre-fetch data for your
        actors _set and _get methods. This helps if you can execute a single
        API call that gets most of the data you need, before any of the actual
        get/set operations take place.
        """
        raise gen.Return()

    @gen.coroutine
    def _get_state(self):
        raise NotImplementedError('_get_state is required for Ensurable')

    @gen.coroutine
    def _set_state(self):
        raise NotImplementedError('_set_state is required for Ensurable')

    @gen.coroutine
    def _ensure(self, option):
        """Compares the desired state with the actual state of a resource.

        Uses the getter for a resource option to determine its current state,
        and then compares it with the desired state. Generally does a simple
        string comparison of the states, but user can optionally define their
        own comparison mechanism as well.

        If the states do not match, then the setter method is called.
        """
        equals = yield self.comparers[option]()

        if equals:
            self.log.debug('Option "%s" matches' % option)
            raise gen.Return()

        self.log.debug('Option "%s" DOES NOT match, calling setter' % option)
        yield self.setters[option]()

    @gen.coroutine
    def _execute(self):
        """A pretty simple execution pipeline for the actor.

        Note: An OrderedDict can be used instead of a plain dict when order
        actually matters for the option setting.
        """
        yield self._precache()

        yield self._ensure('state')

        if self.option('state') == 'absent':
            raise gen.Return()

        for option in self._ensurable_options:
            # We've already managed state .. so make sure we skip the state
            # option and only manage the others.
            if option != 'state':
                yield self._ensure(option)


class HTTPBaseActor(BaseActor):

    """Abstract base class for an HTTP-client based Actor object.

    This class provides common methods for getting access to asynchronous
    HTTP clients, wrapping the executions in appropriate try/except blocks,
    timeouts, etc.

    If you're writing an Actor that uses a remote REST API, this is the
    base class you should subclass from.
    """

    headers = None

    def _get_http_client(self):
        """Returns an asynchronous web client object

        The object is actually of type SimpleAsyncHTTPClient
        """
        return httpclient.AsyncHTTPClient()

    def _get_method(self, post):
        """Returns the appropriate HTTP Method based on the supplied Post data.

        Args:
            post: The post body you intend to submit in the URL request

        Returns:
            'GET' or 'POST'
        """
        # If there is no post data, set the request method to GET
        if post is not None:
            return 'POST'
        else:
            return 'GET'

    def _generate_escaped_url(self, url, args):
        """Takes in a dictionary of arguments and returns a URL line.

        Sorts the arguments so that the returned string is predictable and in
        alphabetical order. Effectively wraps the tornado.httputil.url_concat
        method and properly strips out None values, as well as lowercases
        Bool values.

        Args:
            url: (Str) The URL to append the arguments to
            args: (Dict) Key/Value arguments. Values should be primitives.

        Returns:
            A URL encoded string like this: <url>?foo=bar&abc=xyz
        """

        # Remove keys from the arguments where the value is None
        args = dict((k, v) for k, v in args.items() if v)

        # Convert all Bool values to lowercase strings
        for key, value in args.items():
            if type(value) is bool:
                args[key] = str(value).lower()

        # Now generate the URL
        full_url = httputil.url_concat(url, sorted(args.items()))
        self.log.debug('Generated URL: %s' % full_url)

        return full_url

    # TODO: Add a retry/backoff timer here. If the remote endpoint returns
    # garbled data (ie, maybe a 500 errror or something else thats not in
    # JSON format, we should back off and try again.
    @gen.coroutine
    def _fetch(self, url, post=None, auth_username=None, auth_password=None):
        """Executes a web request asynchronously and yields the body.

        Args:
            url: (Str) The full url path of the API call
            post: (Str) POST body data to submit (if any)
            auth_username: (str) HTTP auth username
            auth_password: (str) HTTP auth password
        """

        # Generate the full request URL and log out what we're doing...
        self.log.debug('Making HTTP request to %s with data: %s' % (url, post))

        # Create the http_request object
        http_client = self._get_http_client()
        http_request = httpclient.HTTPRequest(
            url=url,
            method=self._get_method(post),
            body=post,
            headers=self.headers,
            auth_username=auth_username,
            auth_password=auth_password,
            follow_redirects=True,
            max_redirects=10)

        # Execute the request and raise any exception. Exceptions are not
        # caught here because they are unique to the API endpoints, and thus
        # should be handled by the individual Actor that called this method.
        http_response = yield http_client.fetch(http_request)

        try:
            body = json.loads(http_response.body)
        except ValueError as e:
            raise exceptions.UnparseableResponseFromEndpoint(
                'Unable to parse response from remote API as JSON: %s' % e)

        # Receive a successful return
        raise gen.Return(body)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.rightscale.settings`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Common settings used by many of the `kingpin.actors.rightscale` modules.
"""


import logging
import requests

__author__ = 'Matt Wise <matt@nextdoor.com>'

log = logging.getLogger(__name__)


# Common Settings for the retrying.retry() decorator
#
# Use like this: @retrying.retry(**settings.RETRYING_SETTINGS)
#
def is_retriable_exception(exception):
    """Return true if this RightScale exception is transient.

    Example:
        >>> @retry(retry_on_exception=is_retriable_exception)
    """
    not_retry_codes = ('422',)

    # Only handle requests.exceptions.HTTPError exceptions
    if not isinstance(exception, requests.exceptions.HTTPError):
        return False

    log.debug('Comparing "%s" to "%s".' % (str(exception), not_retry_codes))
    return not any(code in str(exception) for code in not_retry_codes)


RETRYING_SETTINGS = {
    # Verify if we need to retry with the is_retriable_exception
    # method described above.
    'retry_on_exception': is_retriable_exception,

    # Wait up to 10 times
    'stop_max_attempt_number': 10,

    # Add 250ms of random jitter to every retry
    'wait_jitter_max': 250,

    # Add 250ms of sleep to every retry
    'wait_fixed': 250,

    # Now add between 250-2000ms to every retry
    'wait_random_min': 250,
    'wait_random_max': 2000,

    # Finally, add in an exponential backoff timer with a 10s limit
    'wait_exponential_multiplier': 100,
    'wait_exponential_max': 30000
}



from tornado import gen

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'


def mock_tornado(value=None, exc=None):
    """Creates a mock for a coroutine function that returns `value`"""

    @gen.coroutine
    def call(*args, **kwargs):
        call._call_count = call._call_count + 1
        if exc:
            raise exc
        raise gen.Return(value)

    call._call_count = 0
    return call


@gen.coroutine
def tornado_value(value=None):
    """Convers whatever is passed in to a tornado value."""
    raise gen.Return(value)



"""Simple integration tests for the AWS Base."""

from nose.plugins.attrib import attr
import logging

from tornado import testing

from kingpin.actors import exceptions
from kingpin.actors.aws import base
from kingpin.actors.aws import settings
import importlib

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'

log = logging.getLogger(__name__)
logging.getLogger('boto').setLevel(logging.INFO)


class IntegrationBase(testing.AsyncTestCase):
    """High level AWS Base testing."""

    integration = True

    region = 'us-east-1'
    elb_name = 'kingpin-integration-test'

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_01a_check_credentials(self):

        settings.AWS_ACCESS_KEY_ID = 'fake'
        settings.AWS_SECRET_ACCESS_KEY = 'fake'
        actor = base.AWSBaseActor('Test', {'region': self.region})

        # Executing a random function call that is wrapped in _retry.
        # Credentials should fail before "ELB not found" should be raised.
        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor._find_elb('unit-test')

        importlib.reload(settings)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_02a_find_elb(self):

        actor = base.AWSBaseActor('Test', {'region': self.region})

        LB = yield actor._find_elb(self.elb_name)

        self.assertEqual(LB.name, self.elb_name)



"""Simple integration tests for the AWS S3 actors."""

from nose.plugins.attrib import attr
import logging

from tornado import testing

# from kingpin import utils
# from kingpin.actors import exceptions
from kingpin.actors.aws import s3

__author__ = 'Matt Wise <matt@nextdoor.com>'

log = logging.getLogger(__name__)
logging.getLogger('boto').setLevel(logging.INFO)


class IntegrationS3(testing.AsyncTestCase):

    """High level S3 Actor testing.

    These tests will check two things:
    * Creation of S3 buckets with a variety of options
    * Updating these S3 buckets with new parameters
    * Destroying the S3 buckets

    Requirements:
        You have to create an S3 Bucket named kingpin-integration-test and
        place it in the specified region (default us-east-1).
        As with other tests, environment variables AWS_ACCESS_KEY_ID and
        AWS_SECRET_ACCESS_KEY are expected, and the key should have
        permissions to read S3 bucket information.

    Note, these tests must be run in-order. The order is defined by
    their definition order in this file. Nose follows this order according
    to its documentation:

        http://nose.readthedocs.org/en/latest/writing_tests.html
    """

    integration = True

    bucket_name = 'kingpin-integration-test'
    region = 'us-east-1'

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_01_create_bucket(self):
        actor = s3.Bucket(
            options={
                'name': self.bucket_name,
                'region': self.region,
                'state': 'present',
                'logging': {'target': ''},
                'versioning': False
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_02a_set_bucket_policy(self):
        actor = s3.Bucket(
            options={
                'name': self.bucket_name,
                'region': self.region,
                'state': 'present',
                'policy': 'examples/aws.s3/amazon_put.json',
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_02b_delete_bucket_policy(self):
        actor = s3.Bucket(
            options={
                'name': self.bucket_name,
                'region': self.region,
                'state': 'present',
                'policy': '',
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_03a_enable_versioning(self):
        actor = s3.Bucket(
            options={
                'name': self.bucket_name,
                'region': self.region,
                'state': 'present',
                'versioning': True,
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_03b_disable_versioning(self):
        actor = s3.Bucket(
            options={
                'name': self.bucket_name,
                'region': self.region,
                'state': 'present',
                'versioning': False,
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_04a_enable_lifecycle_management(self):
        actor = s3.Bucket(
            options={
                'name': self.bucket_name,
                'region': self.region,
                'state': 'present',
                'lifecycle': [{
                    'id': 'test',
                    'prefix': '/',
                    'status': 'Enabled',
                    'expiration': 30,
                    'transition': {
                        'days': 10,
                        'storage_class': 'GLACIER'
                    }
                }]
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_04b_update_lifecycle_management(self):
        actor = s3.Bucket(
            options={
                'name': self.bucket_name,
                'region': self.region,
                'state': 'present',
                'lifecycle': [{
                    'id': 'test',
                    'prefix': '/',
                    'status': 'Enabled',
                    'expiration': 180,
                    'transition': {
                        'days': 90,
                        'storage_class': 'STANDARD_IA'
                    }
                }]
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_04c_disable_lifecycle_management(self):
        actor = s3.Bucket(
            options={
                'name': self.bucket_name,
                'region': self.region,
                'state': 'present',
                'lifecycle': []
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_09_delete_bucket(self):
        actor = s3.Bucket(
            options={
                'name': self.bucket_name,
                'region': self.region,
                'state': 'absent',
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)



import io
import logging
import os
import time

from tornado import gen
from tornado import testing
from tornado.testing import unittest
import mock
import rainbow_logging_handler
import requests

from kingpin import exceptions
from kingpin import utils
from kingpin.actors import misc
import importlib


class TestUtils(unittest.TestCase):

    def test_str_to_class(self):
        class_string_name = 'tornado.testing.AsyncTestCase'
        returned_class = utils.str_to_class(class_string_name)
        self.assertEqual(testing.AsyncTestCase, returned_class)

        class_string_name = 'kingpin.actors.misc.Sleep'
        returned_class = utils.str_to_class(class_string_name)
        self.assertEqual(misc.Sleep, returned_class)

    def test_populate_with_env(self):
        tokens = {'UNIT_TEST': 'FOOBAR'}
        string = 'Unit %UNIT_TEST% Test'
        expect = 'Unit FOOBAR Test'
        result = utils.populate_with_tokens(string, tokens)
        self.assertEqual(result, expect)

    def test_populate_with_values_not_default(self):
        tokens = {'UNIT_TEST': 'FOOBAR', 'SECOND_UNIT': 'BARBAR'}
        string = 'Unit %UNIT_TEST|DEFAULT% Test %SECOND_UNIT|DEFAULT2%'
        expect = 'Unit FOOBAR Test BARBAR'
        result = utils.populate_with_tokens(string, tokens)
        self.assertEqual(result, expect)

    def test_populate_with_default(self):
        tokens = {'OTHER': 'TOKEN'}
        string = 'Unit %UNIT_TEST|DEFAULT% Test %SECOND_UNIT|1.2,3;4-5%'
        expect = 'Unit DEFAULT Test 1.2,3;4-5'
        result = utils.populate_with_tokens(string, tokens)
        self.assertEqual(result, expect)

    def test_populate_with_unicode_env(self):
        tokens = {'UNIT_TEST': 'FOOBAR'}
        string = 'Unit %UNIT_TEST% Test'
        expect = 'Unit FOOBAR Test'
        result = utils.populate_with_tokens(string, tokens)
        self.assertEqual(result, expect)

    def test_populate_with_bool(self):
        tokens = {'UNIT_TEST': True}
        string = 'Unit %UNIT_TEST% Test'
        expect = 'Unit True Test'
        result = utils.populate_with_tokens(string, tokens)
        self.assertEqual(result, expect)

    def test_populate_with_bogus_data_OK(self):
        tokens = {'UNIT_TEST': {'foobar': 'bat'}}
        string = 'Unit %UNIT_TEST% Test'
        expect = 'Unit %UNIT_TEST% Test'
        result = utils.populate_with_tokens(string, tokens, strict=False)
        self.assertEqual(result, expect)

    def test_populate_with_env_with_missing_variables(self):
        os.environ['UNIT_TEST'] = 'FOOBAR'
        string = 'Unit %UNIT_TEST% Test %NOTFOUNDVARIABLE%'
        with self.assertRaises(LookupError):
            utils.populate_with_tokens(string, os.environ)

    def test_populate_with_env_with_non_string_tokens(self):
        tokens = {'foo': False}
        string = 'Unit test'
        result = utils.populate_with_tokens(string, tokens)
        self.assertEqual(result, string)

    def test_populate_with_not_strict(self):
        tokens = {'UNIT_TEST': 'FOOBAR'}
        string = 'Unit {UNIT_TEST} {FAIL} Test'
        expect = 'Unit FOOBAR {FAIL} Test'
        result = utils.populate_with_tokens(
            string, tokens,
            left_wrapper='{', right_wrapper='}',
            strict=False)
        self.assertEqual(result, expect)

    def test_populate_with_escape_strict_fail(self):
        tokens = {'UNIT_TEST': 'FOOBAR'}
        string = 'Unit \%UNIT_TEST\% Test %TEST_TWO%'
        with self.assertRaises(LookupError):
            utils.populate_with_tokens(string, tokens, strict=True)

    def test_populate_with_escape_strict(self):
        tokens = {'UNIT_TEST': 'FOOBAR'}
        string = 'Unit \%UNIT_TEST\% Test'
        expect = 'Unit %UNIT_TEST% Test'
        result = utils.populate_with_tokens(string, tokens, strict=True)
        self.assertEquals(result, expect)

    def test_populate_with_escape_non_strict(self):
        tokens = {'UNIT_TEST': 'FOOBAR'}
        string = 'Unit \%UNIT_TEST\% Test'
        expect = 'Unit %UNIT_TEST% Test'
        result = utils.populate_with_tokens(string, tokens, strict=False)
        self.assertEquals(result, expect)

    def test_populate_with_escape_non_strict_no_escape(self):
        tokens = {'UNIT_TEST': 'FOOBAR'}
        string = 'Unit \%UNIT_TEST\% Test'
        expect = 'Unit \%UNIT_TEST\% Test'
        result = utils.populate_with_tokens(string, tokens,
                                            strict=False,
                                            remove_escape_sequence=False)
        self.assertEquals(result, expect)

    def test_convert_script_to_dict(self):
        # Should work with string path to a file
        dirname, filename = os.path.split(os.path.abspath(__file__))
        examples = '%s/../../examples' % dirname
        simple = '%s/simple.json' % examples
        ret = utils.convert_script_to_dict(simple, {})
        self.assertEqual(type(ret), dict)

        # Should work with file instance also
        dirname, filename = os.path.split(os.path.abspath(__file__))
        examples = '%s/../../examples' % dirname
        simple = '%s/simple.json' % examples
        instance = open(simple)
        ret = utils.convert_script_to_dict(instance, {})
        self.assertEqual(type(ret), dict)

        # Should definitly support YAML as well
        dirname, filename = os.path.split(os.path.abspath(__file__))
        examples = '%s/../../examples' % dirname
        simple = '%s/simple.yaml' % examples
        instance = open(simple)
        ret = utils.convert_script_to_dict(instance, {})
        self.assertEqual(type(ret), dict)

    def test_convert_script_to_dict_bad_name(self):
        instance = io.StringIO()  # Empty buffer will fail demjson.
        instance.name = 'Somefile.HAHA'

        with self.assertRaises(exceptions.InvalidScriptName):
            utils.convert_script_to_dict(instance, {})

    def test_convert_script_to_dict_junk(self):
        instance = io.StringIO()
        instance.name = 'Somefile.json'

        with self.assertRaises(exceptions.InvalidScript):
            utils.convert_script_to_dict(instance, {})

        with self.assertRaises(exceptions.InvalidScript):
            utils.convert_script_to_dict('junk data', {})

        instance = io.StringIO()
        instance.name = 'Somefile.yaml'
        instance.write('---bad-yaml')

        with self.assertRaises(exceptions.InvalidScript):
            utils.convert_script_to_dict(instance, {})

    def test_exception_logger(self):
        patch = mock.patch.object(utils.logging, 'getLogger')
        with patch as logger:
            importlib.reload(utils)

            @utils.exception_logger
            def raises_exc():
                raise Exception('Whoa')

            with self.assertRaises(Exception):
                raises_exc()

            self.assertEqual(1, logger().debug.call_count)
            logger().debug.assert_called_with(mock.ANY, exc_info=1)


class TestSetupRootLoggerUtils(unittest.TestCase):

    def setUp(self):
        utils.setup_root_logger()

    def test_setup_root_logger(self):
        # Since we're really checking if loggers get created properly,
        # make sure to wipe out any existing logging handlers on the Root
        # logger object.
        log = logging.getLogger()
        log.handlers = []

        # Default logger is basic
        logger = utils.setup_root_logger()
        self.assertEqual(type(logger.handlers[0]), logging.StreamHandler)
        self.assertEqual(logger.level, logging.WARNING)

    def test_setup_root_logger_color(self):
        # Since we're really checking if loggers get created properly,
        # make sure to wipe out any existing logging handlers on the Root
        # logger object.
        log = logging.getLogger()
        log.handlers = []

        # Color logger is nifty
        logger = utils.setup_root_logger(color=True)
        self.assertEqual(
            type(logger.handlers[0]),
            rainbow_logging_handler.RainbowLoggingHandler)
        self.assertEqual(logger.level, logging.WARNING)

    def test_setup_root_logger_with_level(self):
        # Since we're really checking if loggers get created properly,
        # make sure to wipe out any existing logging handlers on the Root
        # logger object.
        log = logging.getLogger()
        log.handlers = []

        logger = utils.setup_root_logger(level='debug')
        self.assertEqual(logger.level, logging.DEBUG)

    def test_setup_root_logger_with_syslog(self):
        # Since we're really checking if loggers get created properly,
        # make sure to wipe out any existing logging handlers on the Root
        # logger object.
        log = logging.getLogger()
        log.handlers = []

        logger = utils.setup_root_logger(syslog='local0')
        self.assertEqual(type(logger.handlers[0]),
                         logging.handlers.SysLogHandler)
        self.assertEqual(logger.handlers[0].facility, 'local0')

    def test_super_httplib_debug_logging(self):
        logger = utils.super_httplib_debug_logging()
        self.assertEqual(10, logger.level)

    def test_order_dict(self):
        d1 = {'foo': [1, 2, 3]}
        d2 = {'foo': [3, 2, 1]}

        ordered_d1 = utils.order_dict(d1)
        ordered_d2 = utils.order_dict(d2)

        self.assertEqual(ordered_d1, ordered_d2)


class TestCoroutineHelpers(testing.AsyncTestCase):

    @testing.gen_test
    def test_retry_with_backoff(self):

        # Define a method that will fail every time
        @gen.coroutine
        @utils.retry(excs=(requests.exceptions.HTTPError), retries=3)
        def raise_exception():
            raise requests.exceptions.HTTPError('Failed')

        with self.assertRaises(requests.exceptions.HTTPError):
            yield raise_exception()

        # Now a method that works
        @gen.coroutine
        @utils.retry(excs=(requests.exceptions.HTTPError), retries=3)
        def work():
            raise gen.Return(True)

        ret = yield work()
        self.assertEqual(ret, True)

    @testing.gen_test
    def testTornadoSleep(self):
        start = time.time()
        yield utils.tornado_sleep(0.1)
        stop = time.time()
        self.assertTrue(stop - start > 0.1)

    @testing.gen_test
    def test_repeating_log(self):
        logger = mock.Mock()  # used for tracking

        # Repeat this message 10 times per second
        # seconds=0 instructs Tornado to invoke this log on every IO loop
        # Below we yield gen.moment to allow IO loop iterations.
        # We do N+1 loops and check N count.
        logid = utils.create_repeating_log(logger.info, 'test', seconds=0)
        yield gen.moment
        yield gen.moment
        yield gen.moment
        yield gen.moment
        yield gen.moment

        utils.clear_repeating_log(logid)
        self.assertEqual(logger.info.call_count, 4)

        # Let's make sure that we don't keep looping our log message.
        yield gen.moment
        yield gen.moment
        self.assertEqual(logger.info.call_count, 4)

    @testing.gen_test
    def test_diff_dicts(self):
        p1 = {'a': 'a', 'b': 'b'}
        p2 = {'a': 'a', 'c': 'c'}

        self.assertEqual(None, utils.diff_dicts(p1, p1))
        self.assertNotEqual(None, utils.diff_dicts(p1, p2))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc
"""
:mod:`kingpin.actors.exceptions`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

All common Actor exceptions
"""

from kingpin import exceptions


class ActorException(exceptions.KingpinException):

    """Base Kingpin Actor Exception"""


class RecoverableActorFailure(ActorException):

    """Base exception that allows script executions to continue on failure.

    This exception class is used to throw an error when an Actor fails, but
    it was an expected and/or acceptable failure.

    This should be used for exceptions that are somewhat normal ... for
    example, trying to delete a ServerArray thats already gone.
    """


class UnrecoverableActorFailure(ActorException):

    """Base exception for unrecoverable failures.

    This exception class should be used for critical failures that should
    always stop a set of Kingpin actors in-place, regardless of the actors
    `warn_on_failure` setting.

    Examples would be when credentials are incorrect, or an unexpected
    exception is caught and there is no known recovery point.
    """


class ActorTimedOut(RecoverableActorFailure):

    """Raised when an Actor takes too long to execute"""


class InvalidActor(UnrecoverableActorFailure):

    """Raised when an invalid Actor name was supplied"""


class InvalidOptions(UnrecoverableActorFailure):

    """Invalid option arguments passed into the Actor object.

    This can be used both for the actual options dict passed into the actor,
    as well as if a the wrong options were used when connecting to a remote
    API.
    """


class InvalidCredentials(UnrecoverableActorFailure):

    """Invalid or missing credentials required for Actor object."""


class UnparseableResponseFromEndpoint(UnrecoverableActorFailure):

    """Invalid response returned from a remote REST endpoint."""


class BadRequest(RecoverableActorFailure):

    """An action failed due to a HTTP 400 error likely due to bad input. """
